import functools
import json
import logging
import os
from functools import partial
from pathlib import Path
from typing import Optional, List

from PySide2 import QtCore, QtWidgets, QtGui
from PySide2.QtCore import QUrl, Signal, Qt
from qasync import asyncSlot

from randovania import VERSION
from randovania.game_description.resources.trick_resource_info import TrickResourceInfo
from randovania.games.game import RandovaniaGame
from randovania.games.patcher_provider import PatcherProvider
from randovania.gui.generated.main_window_ui import Ui_MainWindow
from randovania.gui.lib import common_qt_lib, async_dialog, theme
from randovania.gui.lib.common_qt_lib import open_directory_in_explorer
from randovania.gui.lib.trick_lib import used_tricks, difficulties_for_trick
from randovania.gui.lib.window_manager import WindowManager
from randovania.interface_common import update_checker
from randovania.interface_common.options import Options
from randovania.interface_common.preset_manager import PresetManager
from randovania.layout.base.trick_level import LayoutTrickLevel
from randovania.layout.layout_description import LayoutDescription
from randovania.lib.enum_lib import iterate_enum
from randovania.resolver import debug

_DISABLE_VALIDATION_WARNING = """
<html><head/><body>
<p>While it sometimes throws errors, the validation is what guarantees that your seed is completable.<br/>
Do <span style=" font-weight:600;">not</span> disable if you're uncomfortable with possibly unbeatable seeds.
</p><p align="center">Are you sure you want to disable validation?</p></body></html>
"""


def _t(key: str, disambiguation: Optional[str] = None):
    return QtCore.QCoreApplication.translate("MainWindow", key, disambiguation)


def _update_label_on_show(label: QtWidgets.QLabel, text: str):
    def showEvent(_):
        if label._delayed_text is not None:
            label.setText(label._delayed_text)
            label._delayed_text = None

    label._delayed_text = text
    label.showEvent = showEvent


class MainWindow(WindowManager, Ui_MainWindow):
    newer_version_signal = Signal(str, str)
    options_changed_signal = Signal()
    _is_preview_mode: bool = False

    menu_new_version: Optional[QtWidgets.QAction] = None
    _current_version_url: Optional[str] = None
    _options: Options
    _data_visualizer: Optional[QtWidgets.QWidget] = None
    _map_tracker: QtWidgets.QWidget
    _preset_manager: PresetManager
    GameDetailsSignal = Signal(LayoutDescription)

    InitPostShowSignal = Signal()

    @property
    def _tab_widget(self):
        return self.main_tab_widget

    @property
    def preset_manager(self) -> PresetManager:
        return self._preset_manager

    @property
    def patcher_provider(self) -> PatcherProvider:
        return self._patcher_provider

    @property
    def main_window(self) -> QtWidgets.QMainWindow:
        return self

    @property
    def is_preview_mode(self) -> bool:
        return self._is_preview_mode

    def __init__(self, options: Options, preset_manager: PresetManager,
                 network_client, preview: bool):
        super().__init__()
        self.setupUi(self)
        self.setWindowTitle("Randovania {}".format(VERSION))
        self._is_preview_mode = preview
        self._patcher_provider = PatcherProvider()
        self.setAcceptDrops(True)
        common_qt_lib.set_default_window_icon(self)

        # Remove all hardcoded link color
        about_document: QtGui.QTextDocument = self.about_text_browser.document()
        about_document.setHtml(about_document.toHtml().replace("color:#0000ff;", ""))
        cursor: QtGui.QTextCursor = self.about_text_browser.textCursor()
        cursor.setPosition(0)
        self.about_text_browser.setTextCursor(cursor)

        self.browse_racetime_label.setText(self.browse_racetime_label.text().replace("color:#0000ff;", ""))

        self.intro_label.setText(self.intro_label.text().format(version=VERSION))

        self._preset_manager = preset_manager
        self.network_client = network_client

        if preview:
            debug.set_level(2)

        # Signals
        self.newer_version_signal.connect(self.display_new_version)
        self.options_changed_signal.connect(self.on_options_changed)
        self.GameDetailsSignal.connect(self._open_game_details)
        self.InitPostShowSignal.connect(self.initialize_post_show)

        self.intro_play_now_button.clicked.connect(lambda: self.main_tab_widget.setCurrentWidget(self.tab_play))
        self.open_faq_button.clicked.connect(self._open_faq)
        self.open_database_viewer_button.clicked.connect(partial(self._open_data_visualizer_for_game,
                                                                 RandovaniaGame.METROID_PRIME_ECHOES))

        self.import_permalink_button.clicked.connect(self._import_permalink)
        self.import_game_file_button.clicked.connect(self._import_spoiler_log)
        self.browse_racetime_button.clicked.connect(self._browse_racetime)
        self.create_new_seed_button.clicked.connect(
            lambda: self.main_tab_widget.setCurrentWidget(self.tab_create_seed))

        # Menu Bar
        self.game_menus = []
        self.menu_action_edits = []

        for game in iterate_enum(RandovaniaGame):
            # Sub-Menu in Open Menu
            game_menu = QtWidgets.QMenu(self.menu_open)
            game_menu.setTitle(_t(game.long_name))
            self.menu_open.addAction(game_menu.menuAction())
            self.game_menus.append(game_menu)

            game_trick_details_menu = QtWidgets.QMenu(game_menu)
            game_trick_details_menu.setTitle(_t("Trick Details"))
            self._setup_trick_difficulties_menu_on_show(game_trick_details_menu, game)

            game_data_visualizer_action = QtWidgets.QAction(game_menu)
            game_data_visualizer_action.setText(_t("Data Visualizer"))
            game_data_visualizer_action.triggered.connect(partial(self._open_data_visualizer_for_game, game))

            game_menu.addAction(game_trick_details_menu.menuAction())
            game_menu.addAction(game_data_visualizer_action)

            # Data Editor
            action = QtWidgets.QAction(self)
            action.setText(_t(game.long_name))
            self.menu_internal.addAction(action)
            action.triggered.connect(partial(self._open_data_editor_for_game, game))
            self.menu_action_edits.append(action)

        self.menu_action_edit_existing_database.triggered.connect(self._open_data_editor_prompt)
        self.menu_action_validate_seed_after.triggered.connect(self._on_validate_seed_change)
        self.menu_action_timeout_generation_after_a_time_limit.triggered.connect(self._on_generate_time_limit_change)
        self.menu_action_dark_mode.triggered.connect(self._on_menu_action_dark_mode)
        self.menu_action_experimental_games.triggered.connect(self._on_menu_action_experimental_games)
        self.menu_action_open_auto_tracker.triggered.connect(self._open_auto_tracker)
        self.menu_action_previously_generated_games.triggered.connect(self._on_menu_action_previously_generated_games)
        self.menu_action_log_files_directory.triggered.connect(self._on_menu_action_log_files_directory)
        self.menu_action_layout_editor.triggered.connect(self._on_menu_action_layout_editor)

        # Setting this event only now, so all options changed trigger only once
        options.on_options_changed = self.options_changed_signal.emit
        self._options = options

        self.main_tab_widget.setCurrentIndex(0)

    def closeEvent(self, event):
        self.generate_seed_tab.stop_background_process()
        super().closeEvent(event)

    def dragEnterEvent(self, event: QtGui.QDragEnterEvent):
        from randovania.layout.preset_migration import VersionedPreset

        valid_extensions = [
            LayoutDescription.file_extension(),
            VersionedPreset.file_extension(),
        ]
        valid_extensions_with_dot = {
            f".{extension}"
            for extension in valid_extensions
        }

        for url in event.mimeData().urls():
            ext = os.path.splitext(url.toLocalFile())[1]
            if ext in valid_extensions_with_dot:
                event.acceptProposedAction()
                return

    def dropEvent(self, event: QtGui.QDropEvent):
        from randovania.layout.preset_migration import VersionedPreset

        for url in event.mimeData().urls():
            path = Path(url.toLocalFile())
            if path.suffix == f".{LayoutDescription.file_extension()}":
                self.open_game_details(LayoutDescription.from_file(path))
                return

            elif path.suffix == f".{VersionedPreset.file_extension()}":
                self.main_tab_widget.setCurrentWidget(self.tab_create_seed)
                self.generate_seed_tab.import_preset_file(path)
                return

    def showEvent(self, event: QtGui.QShowEvent):
        self.InitPostShowSignal.emit()

    # Delayed Initialization
    @asyncSlot()
    async def initialize_post_show(self):
        self.InitPostShowSignal.disconnect(self.initialize_post_show)
        logging.info("Will initialize things in post show")
        await self._initialize_post_show_body()
        logging.info("Finished initializing post show")

    async def _initialize_post_show_body(self):
        logging.info("Will load OnlineInteractions")
        from randovania.gui.main_online_interaction import OnlineInteractions
        logging.info("Creating OnlineInteractions...")
        self.online_interactions = OnlineInteractions(self, self.preset_manager, self.network_client, self,
                                                      self._options)

        logging.info("Will load GenerateSeedTab")
        from randovania.gui.generate_seed_tab import GenerateSeedTab
        logging.info("Creating GenerateSeedTab...")
        self.generate_seed_tab = GenerateSeedTab(self, self, self._options)
        logging.info("Running GenerateSeedTab.setup_ui")
        self.generate_seed_tab.setup_ui()

        # Remove pointless Prime 1/3 help tabs
        self.prime_differences_tab.deleteLater()
        self.corruption_differences_tab.deleteLater()

        # Update hints text
        logging.info("Will _update_hints_text")
        self._update_hints_text()
        logging.info("Will hide hint locations combo")
        self._update_hint_locations()

        logging.info("Will update for modified options")
        with self._options:
            self.on_options_changed()

    def _update_hints_text(self):
        from randovania.gui.lib import hints_text

        for game, widget in [
            (RandovaniaGame.METROID_PRIME, self.prime_hint_item_names_tree_widget),
            (RandovaniaGame.METROID_PRIME_ECHOES, self.echoes_hint_item_names_tree_widget),
            (RandovaniaGame.METROID_PRIME_CORRUPTION, self.corruption_hint_item_names_tree_widget),
        ]:
            hints_text.update_hints_text(game, widget)

    def _update_hint_locations(self):
        from randovania.gui.lib import hints_text
        for game, widget in [
            (RandovaniaGame.METROID_PRIME, self.prime_hint_locations_tree_widget),
            (RandovaniaGame.METROID_PRIME_ECHOES, self.echoes_hint_locations_tree_widget),
            (RandovaniaGame.METROID_PRIME_CORRUPTION, self.corruption_hint_locations_tree_widget),
        ]:
            hints_text.update_hint_locations(game, widget)

        self.prime_hint_locations_tab.deleteLater()
        self.corruption_hint_locations_tab.deleteLater()

    # Generate Seed
    def _open_faq(self):
        self.main_tab_widget.setCurrentWidget(self.games_tab)
        self.games_tab_widget.setCurrentWidget(self.help_echoes_tab)
        self.help_prime_tab_widget.setCurrentWidget(self.prime_faq_tab)
        self.help_echoes_tab_widget.setCurrentWidget(self.echoes_faq_tab)
        self.help_corruption_tab_widget.setCurrentWidget(self.corruption_faq_tab)

    async def generate_seed_from_permalink(self, permalink):
        from randovania.lib.status_update_lib import ProgressUpdateCallable
        from randovania.gui.dialog.background_process_dialog import BackgroundProcessDialog

        def work(progress_update: ProgressUpdateCallable):
            from randovania.interface_common import simplified_patcher
            layout = simplified_patcher.generate_layout(progress_update=progress_update,
                                                        permalink=permalink,
                                                        options=self._options)
            progress_update(f"Success! (Seed hash: {layout.shareable_hash})", 1)
            return layout

        new_layout = await BackgroundProcessDialog.open_for_background_task(work, "Creating a game...")
        self.open_game_details(new_layout)

    @asyncSlot()
    async def _import_permalink(self):
        from randovania.gui.dialog.permalink_dialog import PermalinkDialog
        dialog = PermalinkDialog()
        result = await async_dialog.execute_dialog(dialog)
        if result == QtWidgets.QDialog.Accepted:
            permalink = dialog.get_permalink_from_field()
            await self.generate_seed_from_permalink(permalink)

    def _import_spoiler_log(self):
        json_path = common_qt_lib.prompt_user_for_input_game_log(self)
        if json_path is not None:
            layout = LayoutDescription.from_file(json_path)
            self.open_game_details(layout)

    @asyncSlot()
    async def _browse_racetime(self):
        from randovania.gui.dialog.racetime_browser_dialog import RacetimeBrowserDialog
        dialog = RacetimeBrowserDialog()
        if not await dialog.refresh():
            return
        result = await async_dialog.execute_dialog(dialog)
        if result == QtWidgets.QDialog.Accepted:
            await self.generate_seed_from_permalink(dialog.permalink)

    def open_game_details(self, layout: LayoutDescription):
        self.GameDetailsSignal.emit(layout)

    def _open_game_details(self, layout: LayoutDescription):
        from randovania.gui.seed_details_window import SeedDetailsWindow
        details_window = SeedDetailsWindow(self, self._options)
        details_window.update_layout_description(layout)
        details_window.show()
        self.track_window(details_window)

    # Releases info
    async def request_new_data(self):
        from randovania.interface_common import github_releases_data
        await self._on_releases_data(await github_releases_data.get_releases())

    async def _on_releases_data(self, releases: Optional[List[dict]]):
        import markdown

        current_version = update_checker.strict_current_version()
        last_changelog = self._options.last_changelog_displayed

        all_change_logs, new_change_logs, version_to_display = update_checker.versions_to_display_for_releases(
            current_version, last_changelog, releases)

        if version_to_display is not None:
            self.display_new_version(version_to_display)

        if all_change_logs:
            changelog_tab = QtWidgets.QWidget()
            changelog_tab.setObjectName("changelog_tab")
            changelog_tab_layout = QtWidgets.QVBoxLayout(changelog_tab)
            changelog_tab_layout.setContentsMargins(0, 0, 0, 0)
            changelog_tab_layout.setObjectName("changelog_tab_layout")
            changelog_scroll_area = QtWidgets.QScrollArea(changelog_tab)
            changelog_scroll_area.setWidgetResizable(True)
            changelog_scroll_area.setObjectName("changelog_scroll_area")
            changelog_scroll_contents = QtWidgets.QWidget()
            changelog_scroll_contents.setGeometry(QtCore.QRect(0, 0, 489, 337))
            changelog_scroll_contents.setObjectName("changelog_scroll_contents")
            changelog_scroll_layout = QtWidgets.QVBoxLayout(changelog_scroll_contents)
            changelog_scroll_layout.setObjectName("changelog_scroll_layout")

            for entry in all_change_logs:
                changelog_label = QtWidgets.QLabel(changelog_scroll_contents)
                _update_label_on_show(changelog_label, markdown.markdown(entry))
                changelog_label.setObjectName("changelog_label")
                changelog_label.setWordWrap(True)
                changelog_scroll_layout.addWidget(changelog_label)

            changelog_scroll_area.setWidget(changelog_scroll_contents)
            changelog_tab_layout.addWidget(changelog_scroll_area)
            self.main_tab_widget.addTab(changelog_tab, "Change Log")

        if new_change_logs:
            await async_dialog.message_box(self, QtWidgets.QMessageBox.Information,
                                           "What's new", markdown.markdown("\n".join(new_change_logs)))
            with self._options as options:
                options.last_changelog_displayed = current_version

    def display_new_version(self, version: update_checker.VersionDescription):
        if self.menu_new_version is None:
            self.menu_new_version = QtWidgets.QAction("", self)
            self.menu_new_version.triggered.connect(self.open_version_link)
            self.menu_bar.addAction(self.menu_new_version)

        self.menu_new_version.setText("New version available: {}".format(version.tag_name))
        self._current_version_url = version.html_url

    def open_version_link(self):
        if self._current_version_url is None:
            raise RuntimeError("Called open_version_link, but _current_version_url is None")

        QtGui.QDesktopServices.openUrl(QUrl(self._current_version_url))

    # Options
    def on_options_changed(self):
        self.menu_action_validate_seed_after.setChecked(self._options.advanced_validate_seed_after)
        self.menu_action_timeout_generation_after_a_time_limit.setChecked(
            self._options.advanced_timeout_during_generation)
        self.menu_action_dark_mode.setChecked(self._options.dark_mode)
        self.menu_action_experimental_games.setChecked(self._options.experimental_games)

        self.generate_seed_tab.on_options_changed(self._options)
        theme.set_dark_theme(self._options.dark_mode)

    # Menu Actions
    def _open_data_visualizer_for_game(self, game: RandovaniaGame):
        self.open_data_visualizer_at(None, None, game)

    def open_data_visualizer_at(self,
                                world_name: Optional[str],
                                area_name: Optional[str],
                                game: RandovaniaGame = RandovaniaGame.METROID_PRIME_ECHOES,
                                ):
        from randovania.gui.data_editor import DataEditorWindow
        data_visualizer = DataEditorWindow.open_internal_data(game, False)
        self._data_visualizer = data_visualizer

        if world_name is not None:
            data_visualizer.focus_on_world(world_name)

        if area_name is not None:
            data_visualizer.focus_on_area(area_name)

        self._data_visualizer.show()

    def _open_data_editor_for_game(self, game: RandovaniaGame):
        from randovania.gui.data_editor import DataEditorWindow
        self._data_editor = DataEditorWindow.open_internal_data(game, True)
        self._data_editor.show()

    def _open_data_editor_prompt(self):
        from randovania.gui.data_editor import DataEditorWindow

        database_path = common_qt_lib.prompt_user_for_database_file(self)
        if database_path is None:
            return

        with database_path.open("r") as database_file:
            self._data_editor = DataEditorWindow(json.load(database_file), database_path, False, True)
            self._data_editor.show()

    def open_map_tracker(self, configuration: "Preset"):
        from randovania.gui.tracker_window import TrackerWindow, InvalidLayoutForTracker

        try:
            self._map_tracker = TrackerWindow(self._options.tracker_files_path, configuration)
        except InvalidLayoutForTracker as e:
            QtWidgets.QMessageBox.critical(
                self,
                "Unsupported configuration for Tracker",
                str(e)
            )
            return

        self._map_tracker.show()

    # Difficulties stuff

    def _exec_trick_details(self, popup: "TrickDetailsPopup"):
        self._trick_details_popup = popup
        self._trick_details_popup.setWindowModality(Qt.WindowModal)
        self._trick_details_popup.open()

    def _open_trick_details_popup(self, game, trick: TrickResourceInfo, level: LayoutTrickLevel):
        from randovania.gui.dialog.trick_details_popup import TrickDetailsPopup
        self._exec_trick_details(TrickDetailsPopup(self, self, game, trick, level))

    def _setup_trick_difficulties_menu_on_show(self, menu: QtWidgets.QMenu, game: RandovaniaGame):
        def on_show():
            menu.aboutToShow.disconnect(on_show)
            self._setup_difficulties_menu(game, menu)

        menu.aboutToShow.connect(on_show)

    def _setup_difficulties_menu(self, game: RandovaniaGame, menu: QtWidgets.QMenu):
        from randovania.game_description import default_database
        game = default_database.game_description_for(game)
        tricks_in_use = used_tricks(game)

        menu.clear()
        for trick in sorted(game.resource_database.trick, key=lambda _trick: _trick.long_name):
            if trick not in tricks_in_use:
                continue

            trick_menu = QtWidgets.QMenu(self)
            trick_menu.setTitle(_t(trick.long_name))
            menu.addAction(trick_menu.menuAction())

            used_difficulties = difficulties_for_trick(game, trick)
            for i, trick_level in enumerate(iterate_enum(LayoutTrickLevel)):
                if trick_level in used_difficulties:
                    difficulty_action = QtWidgets.QAction(self)
                    difficulty_action.setText(trick_level.long_name)
                    trick_menu.addAction(difficulty_action)
                    difficulty_action.triggered.connect(
                        functools.partial(self._open_trick_details_popup, game, trick, trick_level))

    # ==========

    @asyncSlot()
    async def _on_validate_seed_change(self):
        old_value = self._options.advanced_validate_seed_after
        new_value = self.menu_action_validate_seed_after.isChecked()

        if old_value and not new_value:
            box = QtWidgets.QMessageBox(self)
            box.setWindowTitle("Disable validation?")
            box.setText(_DISABLE_VALIDATION_WARNING)
            box.setIcon(QtWidgets.QMessageBox.Warning)
            box.setStandardButtons(QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)
            box.setDefaultButton(QtWidgets.QMessageBox.No)
            user_response = await async_dialog.execute_dialog(box)
            if user_response != QtWidgets.QMessageBox.Yes:
                self.menu_action_validate_seed_after.setChecked(True)
                return

        with self._options as options:
            options.advanced_validate_seed_after = new_value

    def _on_generate_time_limit_change(self):
        is_checked = self.menu_action_timeout_generation_after_a_time_limit.isChecked()
        with self._options as options:
            options.advanced_timeout_during_generation = is_checked

    def _on_menu_action_dark_mode(self):
        with self._options as options:
            options.dark_mode = self.menu_action_dark_mode.isChecked()

    def _on_menu_action_experimental_games(self):
        with self._options as options:
            options.experimental_games = self.menu_action_experimental_games.isChecked()

    def _open_auto_tracker(self):
        from randovania.gui.auto_tracker_window import AutoTrackerWindow
        self.auto_tracker_window = AutoTrackerWindow(common_qt_lib.get_game_connection(), self._options)
        self.auto_tracker_window.show()

    def _on_menu_action_previously_generated_games(self):
        path = self._options.game_history_path
        try:
            open_directory_in_explorer(path)

        except OSError:
            box = QtWidgets.QMessageBox(QtWidgets.QMessageBox.Information, "Game History",
                                        f"Previously generated games can be found at:\n{path}",
                                        QtWidgets.QMessageBox.Ok, self)
            box.setTextInteractionFlags(Qt.TextSelectableByMouse)
            box.show()

    def _on_menu_action_log_files_directory(self):
        path = self._options.logs_path
        try:
            open_directory_in_explorer(path)

        except OSError:
            box = QtWidgets.QMessageBox(QtWidgets.QMessageBox.Information, "Logs",
                                        f"Randovania logs can be found at:\n{path}",
                                        QtWidgets.QMessageBox.Ok, self)
            box.setTextInteractionFlags(Qt.TextSelectableByMouse)
            box.show()

    def _on_menu_action_layout_editor(self):
        from randovania.gui.corruption_layout_editor import CorruptionLayoutEditor
        self.corruption_editor = CorruptionLayoutEditor()
        self.corruption_editor.show()



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'preset_location_pool.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_PresetLocationPool(object):
    def setupUi(self, PresetLocationPool):
        if not PresetLocationPool.objectName():
            PresetLocationPool.setObjectName(u"PresetLocationPool")
        PresetLocationPool.resize(505, 463)
        self.centralWidget = QWidget(PresetLocationPool)
        self.centralWidget.setObjectName(u"centralWidget")
        self.centralWidget.setMaximumSize(QSize(16777215, 16777215))
        self.verticalLayout = QVBoxLayout(self.centralWidget)
        self.verticalLayout.setSpacing(6)
        self.verticalLayout.setContentsMargins(11, 11, 11, 11)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(2, 2, 2, 2)
        self.randomization_mode_group = QGroupBox(self.centralWidget)
        self.randomization_mode_group.setObjectName(u"randomization_mode_group")
        self.verticalLayout_2 = QVBoxLayout(self.randomization_mode_group)
        self.verticalLayout_2.setSpacing(6)
        self.verticalLayout_2.setContentsMargins(11, 11, 11, 11)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.check_major_minor = QCheckBox(self.randomization_mode_group)
        self.check_major_minor.setObjectName(u"check_major_minor")

        self.verticalLayout_2.addWidget(self.check_major_minor)

        self.randomization_mode_label = QLabel(self.randomization_mode_group)
        self.randomization_mode_label.setObjectName(u"randomization_mode_label")
        self.randomization_mode_label.setWordWrap(True)

        self.verticalLayout_2.addWidget(self.randomization_mode_label)


        self.verticalLayout.addWidget(self.randomization_mode_group)

        self.locations_scroll_area = QScrollArea(self.centralWidget)
        self.locations_scroll_area.setObjectName(u"locations_scroll_area")
        self.locations_scroll_area.setWidgetResizable(True)
        self.locations_scroll_area_contents = QWidget()
        self.locations_scroll_area_contents.setObjectName(u"locations_scroll_area_contents")
        self.locations_scroll_area_contents.setGeometry(QRect(0, 0, 499, 343))
        self.locations_scroll_area_layout = QVBoxLayout(self.locations_scroll_area_contents)
        self.locations_scroll_area_layout.setSpacing(6)
        self.locations_scroll_area_layout.setContentsMargins(11, 11, 11, 11)
        self.locations_scroll_area_layout.setObjectName(u"locations_scroll_area_layout")
        self.locations_scroll_area_layout.setContentsMargins(0, 0, 0, 0)
        self.locations_scroll_area.setWidget(self.locations_scroll_area_contents)

        self.verticalLayout.addWidget(self.locations_scroll_area)

        PresetLocationPool.setCentralWidget(self.centralWidget)

        self.retranslateUi(PresetLocationPool)

        QMetaObject.connectSlotsByName(PresetLocationPool)
    # setupUi

    def retranslateUi(self, PresetLocationPool):
        PresetLocationPool.setWindowTitle(QCoreApplication.translate("PresetLocationPool", u"Location Pool", None))
        self.randomization_mode_group.setTitle(QCoreApplication.translate("PresetLocationPool", u"Randomization Mode", None))
        self.check_major_minor.setText(QCoreApplication.translate("PresetLocationPool", u"Enable major/minor split", None))
        self.randomization_mode_label.setText(QCoreApplication.translate("PresetLocationPool", u"<html><head/><body><p>If this setting is enabled, major items (i.e., major upgrades, Energy Tanks, Dark Temple Keys, and Energy Transfer Modules) and minor items (i.e, expansions) will be shuffled separately.<br/>Major items in excess of the number of major locations will be placed in minor locations, and vice versa.</p></body></html>", None))
    # retranslateUi




# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'preset_echoes_beam_configuration.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_PresetEchoesBeamConfiguration(object):
    def setupUi(self, PresetEchoesBeamConfiguration):
        if not PresetEchoesBeamConfiguration.objectName():
            PresetEchoesBeamConfiguration.setObjectName(u"PresetEchoesBeamConfiguration")
        PresetEchoesBeamConfiguration.resize(466, 454)
        self.centralWidget = QWidget(PresetEchoesBeamConfiguration)
        self.centralWidget.setObjectName(u"centralWidget")
        self.centralWidget.setMaximumSize(QSize(16777215, 16777215))
        self.beam_configuration_tab_layout = QVBoxLayout(self.centralWidget)
        self.beam_configuration_tab_layout.setSpacing(6)
        self.beam_configuration_tab_layout.setContentsMargins(11, 11, 11, 11)
        self.beam_configuration_tab_layout.setObjectName(u"beam_configuration_tab_layout")
        self.beam_configuration_tab_layout.setContentsMargins(4, 4, 4, 0)
        self.beam_configuration_label = QLabel(self.centralWidget)
        self.beam_configuration_label.setObjectName(u"beam_configuration_label")
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.beam_configuration_label.sizePolicy().hasHeightForWidth())
        self.beam_configuration_label.setSizePolicy(sizePolicy)
        self.beam_configuration_label.setAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignTop)
        self.beam_configuration_label.setWordWrap(True)

        self.beam_configuration_tab_layout.addWidget(self.beam_configuration_label)

        self.beam_configuration_group = QGroupBox(self.centralWidget)
        self.beam_configuration_group.setObjectName(u"beam_configuration_group")
        self.beam_configuration_group.setAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignTop)
        self.beam_configuration_group.setFlat(True)
        self.beam_configuration_layout = QGridLayout(self.beam_configuration_group)
        self.beam_configuration_layout.setSpacing(6)
        self.beam_configuration_layout.setContentsMargins(11, 11, 11, 11)
        self.beam_configuration_layout.setObjectName(u"beam_configuration_layout")
        self.beam_configuration_layout.setContentsMargins(0, 0, 0, 0)

        self.beam_configuration_tab_layout.addWidget(self.beam_configuration_group)

        self.beam_configuration_spacer = QSpacerItem(20, 311, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.beam_configuration_tab_layout.addItem(self.beam_configuration_spacer)

        PresetEchoesBeamConfiguration.setCentralWidget(self.centralWidget)

        self.retranslateUi(PresetEchoesBeamConfiguration)

        QMetaObject.connectSlotsByName(PresetEchoesBeamConfiguration)
    # setupUi

    def retranslateUi(self, PresetEchoesBeamConfiguration):
        PresetEchoesBeamConfiguration.setWindowTitle(QCoreApplication.translate("PresetEchoesBeamConfiguration", u"Beam Configuration", None))
        self.beam_configuration_label.setText(QCoreApplication.translate("PresetEchoesBeamConfiguration", u"<html><head/><body><p>Configure what each beam uses as ammo and how much ammo is consumed when shooting it uncharged, charged or with a charge combo.</p><p>Logic always uses the default values.</p><p><span style=\" font-weight:600;\">Known Issue: </span>If you're out of the ammo a beam normally uses, you'll be unable to shoot that beam regardless of what it actually uses to shoot.</p></body></html>", None))
        self.beam_configuration_group.setTitle("")
    # retranslateUi




import dataclasses
import uuid
from typing import Optional, Dict

from PySide2 import QtWidgets, QtGui
from PySide2.QtCore import Qt

from randovania.games.game import RandovaniaGame
from randovania.gui.lib.window_manager import WindowManager
from randovania.lib import enum_lib
from randovania.layout.preset_migration import VersionedPreset, InvalidPreset


class PresetTreeWidget(QtWidgets.QTreeWidget):
    window_manager: WindowManager
    preset_to_item: Dict[uuid.UUID, QtWidgets.QTreeWidgetItem]
    show_experimental: bool = False

    def dropEvent(self, event: QtGui.QDropEvent) -> None:
        item: QtWidgets.QTreeWidgetItem = self.itemAt(event.pos())
        if not item:
            return event.setDropAction(Qt.IgnoreAction)

        source = self.preset_for_item(self.currentItem())
        target = self.preset_for_item(item)

        if source is None or target is None:
            return event.setDropAction(Qt.IgnoreAction)

        if source.game != target.game or source.base_preset_uuid is None:
            return event.setDropAction(Qt.IgnoreAction)

        try:
            source_preset = source.get_preset()
        except InvalidPreset:
            return event.setDropAction(Qt.IgnoreAction)

        self.window_manager.preset_manager.add_new_preset(VersionedPreset.with_preset(
            dataclasses.replace(source_preset, base_preset_uuid=target.uuid)
        ))

        return super().dropEvent(event)

    def preset_for_item(self, item: QtWidgets.QTreeWidgetItem) -> Optional[VersionedPreset]:
        return self.window_manager.preset_manager.preset_for_uuid(item.data(0, Qt.UserRole))

    @property
    def current_preset_data(self) -> Optional[VersionedPreset]:
        for item in self.selectedItems():
            return self.preset_for_item(item)

    def set_show_experimental(self, show_experimental: bool):
        old = self.show_experimental
        self.show_experimental = show_experimental
        if old != show_experimental:
            self.update_items()

    def update_items(self):
        self.clear()

        tree_item = {}
        for game in enum_lib.iterate_enum(RandovaniaGame):
            if game.is_experimental and not self.show_experimental:
                continue

            root = QtWidgets.QTreeWidgetItem(self)
            root.setText(0, game.long_name)
            root.setExpanded(True)
            tree_item[game] = root

        self.preset_to_item = {}

        # Included presets
        for preset in self.window_manager.preset_manager.included_presets.values():
            if preset.game.is_experimental and not self.show_experimental:
                continue

            item = QtWidgets.QTreeWidgetItem(tree_item[preset.game])
            item.setText(0, preset.name)
            item.setExpanded(True)
            item.setData(0, Qt.UserRole, preset.uuid)
            self.preset_to_item[preset.uuid] = item

        # Custom Presets
        for preset in self.window_manager.preset_manager.custom_presets.values():
            if preset.game.is_experimental and not self.show_experimental:
                continue

            item = QtWidgets.QTreeWidgetItem(tree_item[preset.game])
            item.setText(0, preset.name)
            item.setData(0, Qt.UserRole, preset.uuid)
            self.preset_to_item[preset.uuid] = item

        # Set parents after, so don't have issues with order
        for preset in sorted(self.window_manager.preset_manager.custom_presets.values(), key=lambda it: it.name):
            if preset.base_preset_uuid in self.preset_to_item:
                tree_item[preset.game].removeChild(self.preset_to_item[preset.uuid])
                self.preset_to_item[preset.base_preset_uuid].addChild(self.preset_to_item[preset.uuid])

    def select_preset(self, preset: VersionedPreset):
        if preset.uuid in self.preset_to_item:
            self.setCurrentItem(self.preset_to_item[preset.uuid])



from typing import Dict

from randovania.game_description import default_database
from randovania.game_description.game_patches import GamePatches
from randovania.game_description.hint import HintDarkTemple
from randovania.game_description.world.node import PickupNode
from randovania.game_description.resources import resource_info
from randovania.game_description.world.world_list import WorldList
from randovania.games.game import RandovaniaGame
from randovania.games.prime import echoes_items
from randovania.games.prime.patcher_file_lib import hint_lib, guaranteed_item_hint


def create_temple_key_hint(all_patches: Dict[int, GamePatches],
                           player_index: int,
                           temple: HintDarkTemple,
                           area_namers: Dict[int, hint_lib.AreaNamer],
                           ) -> str:
    """
    Creates the text for .
    :param all_patches:
    :param player_index:
    :param temple:
    :param area_namers:
    :return:
    """
    all_world_names = set()

    _TEMPLE_NAMES = ["Dark Agon Temple", "Dark Torvus Temple", "Hive Temple"]
    temple_index = [HintDarkTemple.AGON_WASTES, HintDarkTemple.TORVUS_BOG,
                    HintDarkTemple.SANCTUARY_FORTRESS].index(temple)

    db = default_database.resource_database_for(RandovaniaGame.METROID_PRIME_ECHOES)
    items = [db.get_item(index) for index in echoes_items.DARK_TEMPLE_KEY_ITEMS[temple_index]]

    locations_for_items = guaranteed_item_hint.find_locations_that_gives_items(items, all_patches, player_index)

    for options in locations_for_items.values():
        for player, location in options:
            all_world_names.add(area_namers[player].location_name(location, hide_area=True, color=None))
            break

    temple_name = hint_lib.color_text(hint_lib.TextColor.ITEM, _TEMPLE_NAMES[temple_index])
    names_sorted = [hint_lib.color_text(hint_lib.TextColor.LOCATION, world) for world in sorted(all_world_names)]
    if len(names_sorted) == 0:
        return f"The keys to {temple_name} are nowhere to be found."
    elif len(names_sorted) == 1:
        return f"The keys to {temple_name} can all be found in {names_sorted[0]}."
    else:
        last = names_sorted.pop()
        front = ", ".join(names_sorted)
        return f"The keys to {temple_name} can be found in {front} and {last}."



import dataclasses
import typing
from pathlib import Path

from randovania.bitpacking.json_dataclass import JsonDataclass
from randovania.dol_patching.dol_file import DolFile
from randovania.dol_patching.dol_version import find_version_for_dol
from randovania.games.prime import echoes_dol_versions, all_prime_dol_patches, echoes_dol_patches
from randovania.games.prime.echoes_dol_patches import EchoesDolVersion, BeamConfiguration
from randovania.layout.prime2.echoes_user_preferences import EchoesUserPreferences


@dataclasses.dataclass(frozen=True)
class EchoesDolPatchesData(JsonDataclass):
    energy_per_tank: int
    beam_configuration: BeamConfiguration
    safe_zone_heal_per_second: float
    user_preferences: EchoesUserPreferences
    default_items: dict
    unvisited_room_names: bool
    teleporter_sounds: bool
    dangerous_energy_tank: bool


def apply_patches(game_root: Path, patches_data: EchoesDolPatchesData):
    dol_file = DolFile(_get_dol_path(game_root))

    version = typing.cast(EchoesDolVersion, find_version_for_dol(dol_file, echoes_dol_versions.ALL_VERSIONS))

    dol_file.set_editable(True)
    with dol_file:
        all_prime_dol_patches.apply_remote_execution_patch(version.string_display, dol_file)
        all_prime_dol_patches.apply_energy_tank_capacity_patch(version.health_capacity, patches_data.energy_per_tank,
                                                               dol_file)
        all_prime_dol_patches.apply_reverse_energy_tank_heal_patch(version.sda2_base, version.dangerous_energy_tank,
                                                                   patches_data.dangerous_energy_tank,
                                                                   version.game, dol_file)

        echoes_dol_patches.apply_fixes(version, dol_file)
        echoes_dol_patches.apply_unvisited_room_names(version, dol_file, patches_data.unvisited_room_names)
        echoes_dol_patches.apply_teleporter_sounds(version, dol_file, patches_data.teleporter_sounds)

        echoes_dol_patches.apply_game_options_patch(version.game_options_constructor_address,
                                                    patches_data.user_preferences, dol_file)
        echoes_dol_patches.apply_beam_cost_patch(version.beam_cost_addresses, patches_data.beam_configuration,
                                                 dol_file)
        echoes_dol_patches.apply_safe_zone_heal_patch(version.safe_zone, version.sda2_base,
                                                      patches_data.safe_zone_heal_per_second, dol_file)
        echoes_dol_patches.apply_starting_visor_patch(version.starting_beam_visor, patches_data.default_items,
                                                      dol_file)


def _get_dol_path(game_root: Path) -> Path:
    return game_root.joinpath("sys/main.dol")



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'preset_prime_goal.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_PresetPrimeGoal(object):
    def setupUi(self, PresetPrimeGoal):
        if not PresetPrimeGoal.objectName():
            PresetPrimeGoal.setObjectName(u"PresetPrimeGoal")
        PresetPrimeGoal.resize(383, 329)
        self.centralWidget = QWidget(PresetPrimeGoal)
        self.centralWidget.setObjectName(u"centralWidget")
        self.centralWidget.setMaximumSize(QSize(16777215, 16777215))
        self.goal_layout = QVBoxLayout(self.centralWidget)
        self.goal_layout.setSpacing(6)
        self.goal_layout.setContentsMargins(11, 11, 11, 11)
        self.goal_layout.setObjectName(u"goal_layout")
        self.goal_layout.setContentsMargins(4, 8, 4, 0)
        self.goal_description = QLabel(self.centralWidget)
        self.goal_description.setObjectName(u"goal_description")
        self.goal_description.setWordWrap(True)

        self.goal_layout.addWidget(self.goal_description)

        self.slider_layout = QHBoxLayout()
        self.slider_layout.setSpacing(6)
        self.slider_layout.setObjectName(u"slider_layout")
        self.slider = QSlider(self.centralWidget)
        self.slider.setObjectName(u"slider")
        self.slider.setMaximum(12)
        self.slider.setPageStep(2)
        self.slider.setOrientation(Qt.Horizontal)
        self.slider.setTickPosition(QSlider.TicksBelow)

        self.slider_layout.addWidget(self.slider)

        self.slider_label = QLabel(self.centralWidget)
        self.slider_label.setObjectName(u"slider_label")
        sizePolicy = QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.slider_label.sizePolicy().hasHeightForWidth())
        self.slider_label.setSizePolicy(sizePolicy)
        self.slider_label.setMinimumSize(QSize(20, 0))
        self.slider_label.setAlignment(Qt.AlignCenter)

        self.slider_layout.addWidget(self.slider_label)


        self.goal_layout.addLayout(self.slider_layout)

        self.min_progression_label = QLabel(self.centralWidget)
        self.min_progression_label.setObjectName(u"min_progression_label")

        self.goal_layout.addWidget(self.min_progression_label)

        self.min_progression_spin = QSpinBox(self.centralWidget)
        self.min_progression_spin.setObjectName(u"min_progression_spin")

        self.goal_layout.addWidget(self.min_progression_spin)

        PresetPrimeGoal.setCentralWidget(self.centralWidget)

        self.retranslateUi(PresetPrimeGoal)

        QMetaObject.connectSlotsByName(PresetPrimeGoal)
    # setupUi

    def retranslateUi(self, PresetPrimeGoal):
        PresetPrimeGoal.setWindowTitle(QCoreApplication.translate("PresetPrimeGoal", u"Goal", None))
        self.goal_description.setText(QCoreApplication.translate("PresetPrimeGoal", u"<html><head/><body><p>Controls how many Artifacts will be placed.</p><p>You can always check Artifact Temple for hints where the artifacts were placed.</p></body></html>", None))
        self.slider_label.setText(QCoreApplication.translate("PresetPrimeGoal", u"0", None))
        self.min_progression_label.setText(QCoreApplication.translate("PresetPrimeGoal", u"Only place artifacts after this many actions were performed by the generator.", None))
    # retranslateUi




import copy
from typing import Iterable, List, Dict

from randovania.dol_patching.assembler.ppc import Instruction, BaseInstruction


def assemble_instructions(address: int, instructions: List[BaseInstruction], symbols: Dict[str, int] = None) -> Iterable[int]:
    if symbols is not None:
        symbols = copy.copy(symbols)
    else:
        symbols = {}

    b = address
    for instruction in instructions:
        if instruction.label is not None:
            symbols[instruction.label] = b
        b += instruction.byte_count

    for i, instruction in enumerate(instructions):
        data = list(instruction.bytes_for(address, symbols=symbols))
        yield from data
        address += len(data)


def byte_count(instructions: Iterable[BaseInstruction]) -> int:
    return sum(instruction.byte_count for instruction in instructions)



from unittest.mock import MagicMock

from randovania.games.game import RandovaniaGame
from randovania.gui import game_specific_gui
from randovania.interface_common.preset_editor import PresetEditor


def test_preset_editor_tabs_for(skip_qtbot, game_enum: RandovaniaGame, preset_manager):
    preset = preset_manager.default_preset_for_game(game_enum)
    editor = PresetEditor(preset.get_preset().fork())
    window_manager = MagicMock()

    # Run
    result = game_specific_gui.preset_editor_tabs_for(editor, window_manager)

    # Assert
    assert len(result) >= 3



import asyncio
import locale
import logging.handlers
import os
import sys
import traceback
import typing
from argparse import ArgumentParser
from pathlib import Path

from PySide2 import QtCore, QtWidgets

import randovania

logger = logging.getLogger(__name__)


def display_exception(val: Exception):
    if not isinstance(val, KeyboardInterrupt):
        logging.exception("unhandled exception", exc_info=val)
        box = QtWidgets.QMessageBox(QtWidgets.QMessageBox.Critical,
                                    "An exception was raised",
                                    "An unhandled Exception occurred:\n{}".format(val),
                                    QtWidgets.QMessageBox.Ok)
        from randovania.gui.lib import common_qt_lib
        common_qt_lib.set_default_window_icon(box)
        if val.__traceback__ is not None:
            box.setDetailedText("".join(traceback.format_tb(val.__traceback__)))
        box.exec_()


def catch_exceptions(t, val, tb):
    display_exception(val)


def catch_exceptions_async(loop, context):
    if 'future' in context:
        future: asyncio.Future = context['future']
        logger.exception(context["message"], exc_info=future.exception())
    else:
        logger.critical(str(context))


async def show_main_window(app: QtWidgets.QApplication, options, is_preview: bool):
    from randovania.interface_common.preset_manager import PresetManager
    from randovania.interface_common.options import Options
    options = typing.cast(Options, options)
    preset_manager = PresetManager(options.presets_path)

    logger.info("Loading user presets...")
    await preset_manager.load_user_presets()
    logger.info("Finished loading presets!")

    from randovania.gui.main_window import MainWindow
    logger.info("Preparing main window...")
    main_window = MainWindow(options, preset_manager, app.network_client, is_preview)
    app.main_window = main_window

    logger.info("Displaying main window")
    main_window.show()
    await main_window.request_new_data()


async def show_tracker(app: QtWidgets.QApplication):
    from randovania.gui.auto_tracker_window import AutoTrackerWindow
    options = await _load_options()
    if options is None:
        app.exit(1)
        return

    app.tracker = AutoTrackerWindow(app.game_connection, options)
    logger.info("Displaying auto tracker")
    app.tracker.show()


def show_game_details(app: QtWidgets.QApplication, options, game: Path):
    from randovania.layout.layout_description import LayoutDescription
    from randovania.gui.seed_details_window import SeedDetailsWindow

    layout = LayoutDescription.from_file(game)
    details_window = SeedDetailsWindow(None, options)
    details_window.update_layout_description(layout)
    logger.info("Displaying game details")
    details_window.show()
    app.details_window = details_window


async def show_game_session(app: QtWidgets.QApplication, options, session_id: int):
    from randovania.gui.game_session_window import GameSessionWindow
    from randovania.gui.lib.qt_network_client import QtNetworkClient
    from randovania.interface_common.preset_manager import PresetManager

    network_client: QtNetworkClient = app.network_client

    sessions = [
        session
        for session in await network_client.get_game_session_list()
        if session.id == session_id
    ]
    if not sessions:
        app.quit()
        return
    await network_client.join_game_session(sessions[0], None)

    preset_manager = PresetManager(options.presets_path)

    app.game_session_window = await GameSessionWindow.create_and_update(
        network_client,
        app.game_connection,
        preset_manager,
        None,
        options
    )
    app.game_session_window.show()


async def display_window_for(app, options, command: str, args):
    if command == "tracker":
        await show_tracker(app)
    elif command == "main":
        await show_main_window(app, options, args.preview)
    elif command == "game":
        show_game_details(app, options, args.rdvgame)
    elif command == "session":
        await show_game_session(app, options, args.session_id)
    else:
        raise RuntimeError(f"Unknown command: {command}")


def create_memory_executor(debug_game_backend: bool, options):
    from randovania.interface_common.options import Options
    options = typing.cast(Options, options)

    logger.info("Preparing game backend...")
    if debug_game_backend:
        from randovania.gui.debug_backend_window import DebugExecutorWindow
        backend = DebugExecutorWindow()
        backend.show()
    else:
        try:
            from randovania.game_connection.executor.dolphin_executor import DolphinExecutor
        except ImportError as e:
            from randovania.gui.lib import common_qt_lib
            import traceback
            common_qt_lib.show_install_visual_cpp_redist("{}\n\nTraceback:\n{}".format(
                e,
                "\n".join(traceback.format_tb(e.__traceback__)))
            )
            raise SystemExit(1)

        from randovania.game_connection.executor.nintendont_executor import NintendontExecutor
        from randovania.game_connection.memory_executor_choice import MemoryExecutorChoice

        logger.info("Loaded all game backends...")
        if options.game_backend == MemoryExecutorChoice.NINTENDONT and options.nintendont_ip is not None:
            backend = NintendontExecutor(options.nintendont_ip)
        else:
            backend = DolphinExecutor()

    logger.info("Game backend configured: %s", type(backend))
    return backend


async def _load_options():
    logger.info("Loading up user preferences code...")
    from randovania.interface_common.options import Options
    from randovania.gui.lib import startup_tools, theme

    logger.info("Restoring saved user preferences...")
    options = Options.with_default_data_dir()
    if not await startup_tools.load_options_from_disk(options):
        return None

    logger.info("Creating user preferences folder")
    import dulwich.repo
    import dulwich.errors
    try:
        dulwich.repo.Repo(options.user_dir)

    except (dulwich.errors.NotGitRepository, ValueError):
        options.user_dir.mkdir(parents=True, exist_ok=True)
        dulwich.repo.Repo.init(options.user_dir)

    theme.set_dark_theme(options.dark_mode)

    from randovania.layout.preset_migration import VersionedPreset
    for old_preset in options.data_dir.joinpath("presets").glob("*.randovania_preset"):
        old_preset.rename(old_preset.with_name(f"{old_preset.stem}.{VersionedPreset.file_extension()}"))

    logger.info("Loaded user preferences")

    return options


def start_logger(data_dir: Path, is_preview: bool):
    # Ensure the log dir exists early on
    log_dir = data_dir.joinpath("logs")
    log_dir.mkdir(parents=True, exist_ok=True)

    randovania.setup_logging('DEBUG' if is_preview else 'INFO', log_dir.joinpath(f"logger.log"))


def create_loop(app: QtWidgets.QApplication) -> asyncio.AbstractEventLoop:
    os.environ['QT_API'] = "PySide2"
    import qasync
    loop: asyncio.AbstractEventLoop = qasync.QEventLoop(app)
    asyncio.set_event_loop(loop)

    sys.excepthook = catch_exceptions
    loop.set_exception_handler(catch_exceptions_async)
    return loop


async def qt_main(app: QtWidgets.QApplication, data_dir: Path, args):
    app.setQuitOnLastWindowClosed(False)
    app.network_client = None
    logging.info("Loading server client...")
    from randovania.gui.lib.qt_network_client import QtNetworkClient
    logging.info("Configuring server client...")
    app.network_client = QtNetworkClient(data_dir)
    logging.info("Server client ready.")

    if args.login_as_guest:
        logging.info("Logging as %s", args.login_as_guest)
        await app.network_client.login_as_guest(args.login_as_guest)

    options = await _load_options()
    if options is None:
        app.exit(1)
        return

    executor = create_memory_executor(args.debug_game_backend, options)

    logging.info("Configuring game connection with the backend...")
    from randovania.game_connection.game_connection import GameConnection
    app.game_connection = GameConnection(executor)

    logging.info("Configuring qasync...")
    import qasync

    @qasync.asyncClose
    async def _on_last_window_closed():
        await app.network_client.disconnect_from_server()
        await app.game_connection.stop()
        logger.info("Last QT window closed")

    app.setQuitOnLastWindowClosed(True)
    app.lastWindowClosed.connect(_on_last_window_closed, QtCore.Qt.QueuedConnection)

    await asyncio.gather(app.game_connection.start(),
                         display_window_for(app, options, args.command, args))


def run(args):
    locale.setlocale(locale.LC_ALL, "")  # use system's default locale
    QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling, True)

    data_dir = args.custom_network_storage
    if data_dir is None:
        from randovania.interface_common import persistence
        data_dir = persistence.local_data_dir()

    is_preview = args.preview
    start_logger(data_dir, is_preview)
    app = QtWidgets.QApplication(sys.argv)

    def main_done(done: asyncio.Task):
        e: typing.Optional[Exception] = done.exception()
        if e is not None:
            display_exception(e)
            app.exit(1)

    loop = create_loop(app)
    with loop:
        loop.create_task(qt_main(app, data_dir, args)).add_done_callback(main_done)
        loop.run_forever()


def create_subparsers(sub_parsers):
    parser: ArgumentParser = sub_parsers.add_parser(
        "gui",
        help="Run the Graphical User Interface"
    )
    parser.add_argument("--preview", action="store_true", help="Activates preview features")
    parser.add_argument("--custom-network-storage", type=Path, help="Use a custom path to store the network login.")
    parser.add_argument("--login-as-guest", type=str, help="Login as the given quest user")
    parser.add_argument("--debug-game-backend", action="store_true", help="Opens the debug game backend.")

    gui_parsers = parser.add_subparsers(dest="command")
    gui_parsers.add_parser("main", help="Displays the Main Window").set_defaults(func=run)
    gui_parsers.add_parser("tracker", help="Opens only the auto tracker").set_defaults(func=run)

    game_parser = gui_parsers.add_parser("game", help="Opens an rdvgame")
    game_parser.add_argument("rdvgame", type=Path, help="Path ")
    game_parser.set_defaults(func=run)

    session_parser = gui_parsers.add_parser("session", help="Connects to a game session")
    session_parser.add_argument("session_id", type=int, help="Id of the session")
    session_parser.set_defaults(func=run)

    def check_command(args):
        if args.command is None:
            parser.print_help()
            raise SystemExit(1)

    parser.set_defaults(func=check_command)



import dataclasses

import typing


def _handle_optional(type_):
    if getattr(type_, "__origin__", None) is typing.Union and (
            len(type_.__args__) == 2 and type_.__args__[1] is type(None)):
        return type_.__args__[0], True
    return type_, False


class DataclassPostInitTypeCheck:
    def __post_init__(self):
        for f in dataclasses.fields(self):
            v = getattr(self, f.name)
            resolved_type, optional = _handle_optional(f.type)
            if optional and v is None:
                continue
            if not isinstance(v, resolved_type):
                raise ValueError(f"Unexpected type for field {f.name} ({v}): Got {type(v)}, expected {f.type}.")



import pytest

from randovania.bitpacking import bitpacking
from randovania.bitpacking.bitpacking import BitPackDecoder
from randovania.games.game import RandovaniaGame
from randovania.layout.base.trick_level import LayoutTrickLevel
from randovania.layout.base.trick_level_configuration import TrickLevelConfiguration


@pytest.fixture(
    params=[
        {"encoded": b'\x00\x00', "bit_count": 15, "json": {"minimal_logic": False, "specific_levels": {}}},
        {"encoded": b'\x80', "bit_count": 1, "json": {"minimal_logic": True, "specific_levels": {}}},
        {"encoded": b'X\x00\x00', "bit_count": 18, "json": {
            "minimal_logic": False, "specific_levels": {"Dash": "expert"}}},
        {"encoded": b'f3\x00\x00', "bit_count": 27, "json": {"minimal_logic": False, "specific_levels": {
            i: "hypermode"
            for i in ["Dash", "BombJump", "Movement", "BSJ"]
        }}},
    ],
    name="trick_level_data")
def _trick_level_data(request, mocker, echoes_game_description):
    tricks = echoes_game_description.resource_database.trick[:14]
    mocker.patch("randovania.layout.base.trick_level_configuration._all_tricks", return_value=tricks)
    return (request.param["encoded"], request.param["bit_count"],
            TrickLevelConfiguration.from_json(request.param["json"], game=RandovaniaGame.METROID_PRIME_ECHOES))


def test_decode(trick_level_data):
    # Setup
    data, _, expected = trick_level_data

    # Run
    decoder = BitPackDecoder(data)
    result = TrickLevelConfiguration.bit_pack_unpack(decoder, {
        "reference": TrickLevelConfiguration(False, {}, RandovaniaGame.METROID_PRIME_ECHOES),
    })

    # Assert
    assert result == expected


def test_encode(trick_level_data):
    expected_bytes, expected_bit_count, value = trick_level_data

    # Run
    result, bit_count = bitpacking.pack_results_and_bit_count(value.bit_pack_encode({}))

    # Assert
    assert result == expected_bytes
    assert bit_count == expected_bit_count


def test_encode_no_tricks_are_removed():
    from_json = TrickLevelConfiguration.from_json({"minimal_logic": False, "specific_levels": {"Dash": "disabled"}},
                                                  game=RandovaniaGame.METROID_PRIME_ECHOES)

    encoded, byte_count = bitpacking.pack_results_and_bit_count(from_json.bit_pack_encode({}))

    assert encoded == b'\x00\x00\x00\x00'
    assert byte_count == 26

    decoder = BitPackDecoder(encoded)
    decoded = TrickLevelConfiguration.bit_pack_unpack(
        decoder, {"reference": TrickLevelConfiguration(False, {}, RandovaniaGame.METROID_PRIME_ECHOES), })

    assert decoded.specific_levels == {}


def test_set_level_for_trick_remove(echoes_resource_database):
    trick = echoes_resource_database.trick[0]
    config = TrickLevelConfiguration(False, {}, RandovaniaGame.METROID_PRIME_ECHOES)

    assert config.level_for_trick(trick) == LayoutTrickLevel.DISABLED

    config = config.set_level_for_trick(trick, LayoutTrickLevel.ADVANCED)
    assert config.level_for_trick(trick) == LayoutTrickLevel.ADVANCED

    config = config.set_level_for_trick(trick, LayoutTrickLevel.DISABLED)
    assert config.level_for_trick(trick) == LayoutTrickLevel.DISABLED


def test_pretty_description_minimal_logic():
    config = TrickLevelConfiguration(True, {}, RandovaniaGame.METROID_PRIME_ECHOES)
    assert config.pretty_description == "Minimal Logic"


@pytest.mark.parametrize(["levels", "expected"], [
    ({}, "All tricks disabled"),
    ({i: LayoutTrickLevel.HYPERMODE for i in ["Dash", "BombJump", "Movement", "BSJ"]},
     "Enabled tricks: 21 at Disabled, 4 at Hypermode"),
    ({"Dash": LayoutTrickLevel.HYPERMODE, "BombJump": LayoutTrickLevel.HYPERMODE,
      "AirUnderwater": LayoutTrickLevel.ADVANCED},
     "Enabled tricks: 22 at Disabled, Air Underwater at Advanced, 2 at Hypermode"),
])
def test_pretty_description_tricks_echoes(levels, expected):
    config = TrickLevelConfiguration(False, levels, RandovaniaGame.METROID_PRIME_ECHOES)
    assert config.pretty_description == expected



from pathlib import Path
from unittest.mock import MagicMock, AsyncMock, ANY, call

import pytest
from discord_slash import ComponentType

from randovania.games.game import RandovaniaGame
from randovania.server.discord.database_command import DatabaseCommandCog, SplitWorld


@pytest.mark.asyncio
async def test_on_ready():
    # Setup
    cog = DatabaseCommandCog({"guild": 1234}, MagicMock())
    slash = cog.bot.slash
    slash.sync_all_commands = AsyncMock()

    # Run
    await cog.on_ready()

    # Assert
    slash.add_slash_command.assert_called_once_with(
        cog.database_command,
        name="database-inspect",
        description="Consult the Randovania's logic database for one specific room.",
        guild_ids=[1234],
        options=[ANY],
    )
    slash.add_component_callback.assert_called_once_with(
        cog.on_database_component,
        components=ANY,
        use_callback_name=False,
    )
    slash.sync_all_commands.assert_awaited_once_with()


@pytest.mark.asyncio
async def test_database_command():
    # Setup
    cog = DatabaseCommandCog({"guild": 1234}, MagicMock())

    area = MagicMock()
    area.name = "The Area"
    cog._split_worlds[RandovaniaGame.METROID_PRIME_CORRUPTION] = [
        SplitWorld(MagicMock(), "The World", [area]),
    ]

    ctx = AsyncMock()

    # Run
    await cog.database_command(ctx, RandovaniaGame.METROID_PRIME_CORRUPTION.value)

    # Assert
    ctx.send.assert_awaited_once_with(
        embed=ANY, components=ANY, hidden=True
    )


@pytest.mark.asyncio
async def test_on_database_world_selected():
    # Setup
    cog = DatabaseCommandCog({"guild": 1234}, MagicMock())

    area = MagicMock()
    area.name = "The Area"
    cog._split_worlds[RandovaniaGame.METROID_PRIME_CORRUPTION] = [
        tuple(),
        SplitWorld(MagicMock(), "The World", [area]),
    ]

    ctx = AsyncMock()
    ctx.selected_options = [f"{RandovaniaGame.METROID_PRIME_CORRUPTION.value}_world_1"]

    # Run
    await cog.on_database_world_selected(ctx, RandovaniaGame.METROID_PRIME_CORRUPTION)

    # Assert
    ctx.edit_origin.assert_awaited_once_with(
        embed=ANY,
        components=[
            ANY,
            {"type": ComponentType.actionrow, "components": ANY},
        ],
    )


@pytest.mark.asyncio
async def test_on_database_area_selected(echoes_game_description, mocker):
    # Setup
    mock_file: MagicMock = mocker.patch("discord.File")
    mock_digraph: MagicMock = mocker.patch("graphviz.Digraph")
    dot: MagicMock = mock_digraph.return_value
    dot.render.return_value = "bar"
    Path("bar").write_bytes(b"1234")

    cog = DatabaseCommandCog({"guild": 1234}, MagicMock())

    ctx = AsyncMock()
    ctx.selected_options = [f"area_1"]

    world = echoes_game_description.world_list.worlds[2]
    area = world.areas[0]
    split_world = SplitWorld(world, "The World", [MagicMock(), area])

    # Run
    await cog.on_database_area_selected(ctx, RandovaniaGame.METROID_PRIME_ECHOES, split_world, 2)

    # Assert
    ctx.send.assert_awaited_once_with(
        content=f"Requested by {ctx.author.display_name}.",
        embed=ANY,
        files=[mock_file.return_value],
        components=[ANY],
    )
    mock_digraph.assert_called_once_with(comment=area.name)
    dot.node.assert_has_calls([
        call(node.name)
        for node in area.nodes
    ])
    dot.render.assert_called_once_with(format="png", cleanup=True)
    mock_file.assert_called_once_with(Path("bar"))
    assert not Path("bar").is_file()


@pytest.mark.asyncio
async def test_on_area_node_selection(echoes_game_description, mocker):
    # Setup
    mock_embed: MagicMock = mocker.patch("discord.Embed")

    cog = DatabaseCommandCog({"guild": 1234}, MagicMock())

    world = echoes_game_description.world_list.worlds[2]
    area = world.areas[2]

    ctx = AsyncMock()
    ctx.selected_options = [area.nodes[0].name, area.nodes[2].name]

    # Run
    await cog.on_area_node_selection(ctx, RandovaniaGame.METROID_PRIME_ECHOES, area)

    # Assert
    ctx.edit_origin.assert_awaited_once_with(embed=mock_embed.return_value)
    mock_embed.assert_called_once_with(title=ctx.origin_message.embeds[0].title)
    mock_embed.return_value.add_field.assert_has_calls([
        call(
            name=area.nodes[0].name,
            value=ANY,
            inline=False,
        ),
        call(
            name=area.nodes[2].name,
            value=ANY,
            inline=False,
        ),
    ])



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'item_configuration_popup.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *

from randovania.gui.lib.scroll_protected import ScrollProtectedSpinBox


class Ui_ItemConfigurationPopup(object):
    def setupUi(self, ItemConfigurationPopup):
        if not ItemConfigurationPopup.objectName():
            ItemConfigurationPopup.setObjectName(u"ItemConfigurationPopup")
        ItemConfigurationPopup.resize(534, 99)
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(ItemConfigurationPopup.sizePolicy().hasHeightForWidth())
        ItemConfigurationPopup.setSizePolicy(sizePolicy)
        ItemConfigurationPopup.setMaximumSize(QSize(16777215, 16777215))
        self.gridLayout_2 = QGridLayout(ItemConfigurationPopup)
        self.gridLayout_2.setSpacing(6)
        self.gridLayout_2.setContentsMargins(11, 11, 11, 11)
        self.gridLayout_2.setObjectName(u"gridLayout_2")
        self.gridLayout_2.setContentsMargins(0, 0, 0, 0)
        self.item_name_label = QLabel(ItemConfigurationPopup)
        self.item_name_label.setObjectName(u"item_name_label")
        self.item_name_label.setMinimumSize(QSize(150, 0))
        font = QFont()
        font.setBold(True)
        font.setWeight(75)
        self.item_name_label.setFont(font)

        self.gridLayout_2.addWidget(self.item_name_label, 1, 0, 1, 1)

        self.shuffled_radio = QRadioButton(ItemConfigurationPopup)
        self.shuffled_radio.setObjectName(u"shuffled_radio")

        self.gridLayout_2.addWidget(self.shuffled_radio, 1, 4, 1, 1)

        self.warning_label = QLabel(ItemConfigurationPopup)
        self.warning_label.setObjectName(u"warning_label")
        self.warning_label.setWordWrap(True)

        self.gridLayout_2.addWidget(self.warning_label, 3, 0, 1, 6)

        self.vanilla_radio = QRadioButton(ItemConfigurationPopup)
        self.vanilla_radio.setObjectName(u"vanilla_radio")

        self.gridLayout_2.addWidget(self.vanilla_radio, 1, 2, 1, 1)

        self.excluded_radio = QRadioButton(ItemConfigurationPopup)
        self.excluded_radio.setObjectName(u"excluded_radio")

        self.gridLayout_2.addWidget(self.excluded_radio, 1, 1, 1, 1)

        self.starting_radio = QRadioButton(ItemConfigurationPopup)
        self.starting_radio.setObjectName(u"starting_radio")

        self.gridLayout_2.addWidget(self.starting_radio, 1, 3, 1, 1)

        self.provided_ammo_label = QLabel(ItemConfigurationPopup)
        self.provided_ammo_label.setObjectName(u"provided_ammo_label")
        self.provided_ammo_label.setAlignment(Qt.AlignRight|Qt.AlignTrailing|Qt.AlignVCenter)
        self.provided_ammo_label.setWordWrap(True)

        self.gridLayout_2.addWidget(self.provided_ammo_label, 2, 0, 1, 4)

        self.provided_ammo_spinbox = ScrollProtectedSpinBox(ItemConfigurationPopup)
        self.provided_ammo_spinbox.setObjectName(u"provided_ammo_spinbox")

        self.gridLayout_2.addWidget(self.provided_ammo_spinbox, 2, 4, 1, 2)

        self.shuffled_spinbox = ScrollProtectedSpinBox(ItemConfigurationPopup)
        self.shuffled_spinbox.setObjectName(u"shuffled_spinbox")
        self.shuffled_spinbox.setMinimum(1)
        self.shuffled_spinbox.setMaximum(99)

        self.gridLayout_2.addWidget(self.shuffled_spinbox, 1, 5, 1, 1)

        self.separator_line = QFrame(ItemConfigurationPopup)
        self.separator_line.setObjectName(u"separator_line")
        self.separator_line.setFrameShape(QFrame.HLine)
        self.separator_line.setFrameShadow(QFrame.Sunken)

        self.gridLayout_2.addWidget(self.separator_line, 0, 0, 1, 6)


        self.retranslateUi(ItemConfigurationPopup)

        QMetaObject.connectSlotsByName(ItemConfigurationPopup)
    # setupUi

    def retranslateUi(self, ItemConfigurationPopup):
        ItemConfigurationPopup.setWindowTitle(QCoreApplication.translate("ItemConfigurationPopup", u"Item Configuration", None))
        self.item_name_label.setText(QCoreApplication.translate("ItemConfigurationPopup", u"Unlimited Beam Ammo", None))
        self.shuffled_radio.setText(QCoreApplication.translate("ItemConfigurationPopup", u"Shuffled", None))
        self.warning_label.setText("")
        self.vanilla_radio.setText(QCoreApplication.translate("ItemConfigurationPopup", u"Vanilla", None))
        self.excluded_radio.setText(QCoreApplication.translate("ItemConfigurationPopup", u"Excluded", None))
        self.starting_radio.setText(QCoreApplication.translate("ItemConfigurationPopup", u"Starting", None))
#if QT_CONFIG(tooltip)
        self.provided_ammo_label.setToolTip(QCoreApplication.translate("ItemConfigurationPopup", u"<html><head/><body><p>When this item is collected, it also gives this amount of the given ammos.</p><p>This is included in the calculation of how much each pickup of this ammo gives.</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.provided_ammo_label.setText(QCoreApplication.translate("ItemConfigurationPopup", u"<html><head/><body><p>Provided Ammo (XXXX and YYYY)</p></body></html>", None))
    # retranslateUi




import struct

import pytest

from randovania.dol_patching.dol_file import DolFile, DolHeader, Section, _NUM_SECTIONS


@pytest.fixture()
def dol_file(tmp_path):
    section_size = 0x1C2
    sections = [Section(0, 0, 0)] * _NUM_SECTIONS
    sections[0] = Section(0x100, base_address=0x2000, size=section_size)

    data = bytearray(b"\x00" * (0x100 + section_size))
    data[0:0x100] = DolHeader(tuple(sections), 0, 0, 0).as_bytes()

    tmp_path.joinpath("test.dol").write_bytes(data)
    dol_file = DolFile(tmp_path.joinpath("test.dol"))
    return dol_file



import pytest

from randovania.game_description.resources.item_resource_info import ItemResourceInfo
from randovania.game_description.resources.pickup_entry import PickupEntry, PickupModel
from randovania.games.game import RandovaniaGame


def test_extra_resources_maximum(generic_item_category):
    item = ItemResourceInfo(0, "Item", "Item", 2, None)
    msg = "Attempt to give 5 of Item, more than max capacity"

    with pytest.raises(ValueError, match=msg):
        PickupEntry(
            name="broken",
            model=PickupModel(RandovaniaGame.METROID_PRIME_ECHOES, "Nothing"),
            item_category=generic_item_category,
            broad_category=generic_item_category,
            progression=(),
            extra_resources=(
                (item, 5),
            )
        )



from random import Random
from typing import Dict

from randovania.game_description.game_patches import GamePatches
from randovania.game_description.hint import HintLocationPrecision
from randovania.game_description.world.node import LogbookNode
from randovania.game_description.world.world_list import WorldList
from randovania.games.prime.patcher_file_lib import hint_lib
from randovania.games.prime.patcher_file_lib.hint_formatters import LocationFormatter, GuardianFormatter, \
    TemplatedFormatter, RelativeAreaFormatter
from randovania.games.prime.patcher_file_lib.hint_lib import create_simple_logbook_hint
from randovania.games.prime.patcher_file_lib.hint_name_creator import LocationHintCreator
from randovania.games.prime.patcher_file_lib.item_hints import RelativeItemFormatter
from randovania.interface_common.players_configuration import PlayersConfiguration

_JOKE_HINTS = [
    # Guidelines for joke hints:
    # 1. They should clearly be jokes, and not real hints or the result of a bug.
    # 2. They shouldn't reference real-world people.
    # 3. They should be understandable by as many people as possible.
    #
    "By this point in your run, you should have consumed at least 200 mL of water to maintain optimum hydration.",
    "Make sure to collect an Energy Transfer Module; otherwise your run won't be valid!",
    "Adam has not yet authorized the use of this hint.",
    "Back in my day, we didn't need hints!",
    "Hear the words of O-Lir, last Sentinel of the Fortress Temple. May they serve you well.",
    "Warning! Dark Aether's atmosphere is dangerous! Energized Safe Zones don't last forever!",
    "A really important item can be found at - (transmission ends)",
    "Did you know that Bigfoot and Santa Claus exist in the Metroid Prime canon?",
    "I hear. Them. Everywhere. They're coming. Can't sleep. Ever. They'll eat me. Eat.",
    "Space Pirates, strangely, dislike theft.",
    "Power Bomb Expansions are just space hamburgers.",
    "While in Morph Ball mode, press &image=SI,0.70,0.68,D523DE3B; to drop Bombs.",
    "Charge your beam to fire a normal shot when out of ammo.",
    "Movement in Morph Ball mode is faster than unmorphed, even without Boost Ball.",
    "While walking, holding L makes you move faster.",
    "I wonder how many players will see this message...?",
    "We've been trying to contact you about your ship's extended warranty.",
    "The Scan Visor is useful for gathering information.",
]


def create_hints(all_patches: Dict[int, GamePatches],
                 players_config: PlayersConfiguration,
                 world_list: WorldList,
                 area_namers: Dict[int, hint_lib.AreaNamer],
                 rng: Random,
                 ) -> list:
    """
    Creates the string patches entries that changes the Lore scans in the game for item pickups
    :param all_patches:
    :param players_config:
    :param world_list:
    :param area_namers:
    :param rng:
    :return:
    """

    hint_name_creator = LocationHintCreator(world_list, area_namers, rng, _JOKE_HINTS)
    this_area_namer = area_namers[players_config.player_index]
    patches = all_patches[players_config.player_index]

    location_formatters: Dict[HintLocationPrecision, LocationFormatter] = {
        HintLocationPrecision.KEYBEARER: TemplatedFormatter(
            "The Flying Ing Cache in {node} contains {determiner}{pickup}.", this_area_namer),
        HintLocationPrecision.GUARDIAN: GuardianFormatter(),
        HintLocationPrecision.LIGHT_SUIT_LOCATION: TemplatedFormatter(
            "U-Mos's reward for returning the Sanctuary energy is {determiner}{pickup}.", this_area_namer),
        HintLocationPrecision.DETAILED: TemplatedFormatter("{determiner.title}{pickup} can be found in {node}.",
                                                           this_area_namer),
        HintLocationPrecision.WORLD_ONLY: TemplatedFormatter("{determiner.title}{pickup} can be found in {node}.",
                                                             this_area_namer),
        HintLocationPrecision.RELATIVE_TO_AREA: RelativeAreaFormatter(world_list, patches),
        HintLocationPrecision.RELATIVE_TO_INDEX: RelativeItemFormatter(world_list, patches, players_config),
    }

    hints_for_asset: Dict[int, str] = {}
    for asset, hint in patches.hints.items():
        hints_for_asset[asset.asset_id] = hint_name_creator.create_message_for_hint(hint, all_patches, players_config,
                                                                                    location_formatters)

    return [
        create_simple_logbook_hint(
            logbook_node.string_asset_id,
            hints_for_asset.get(logbook_node.string_asset_id, "Someone forgot to leave a message."),
        )
        for logbook_node in world_list.all_nodes
        if isinstance(logbook_node, LogbookNode)
    ]


def hide_hints(world_list: WorldList) -> list:
    """
    Creates the string patches entries that changes the Lore scans in the game
    completely useless text.
    :return:
    """

    return [create_simple_logbook_hint(logbook_node.string_asset_id, "Some item was placed somewhere.")
            for logbook_node in world_list.all_nodes if isinstance(logbook_node, LogbookNode)]



import dataclasses
import datetime
import logging
import random
import uuid
from functools import partial
from pathlib import Path
from typing import Optional, Callable

from PySide2 import QtWidgets, QtCore, QtGui
from PySide2.QtCore import QTimer
from qasync import asyncSlot

from randovania.gui.generated.main_window_ui import Ui_MainWindow
from randovania.gui.lib import preset_describer, common_qt_lib, async_dialog
from randovania.gui.lib.background_task_mixin import BackgroundTaskMixin
from randovania.gui.lib.generation_failure_handling import GenerationFailureHandler
from randovania.gui.lib.window_manager import WindowManager
from randovania.gui.preset_settings.logic_settings_window import LogicSettingsWindow
from randovania.interface_common import simplified_patcher
from randovania.interface_common.options import Options
from randovania.interface_common.preset_editor import PresetEditor
from randovania.layout.layout_description import LayoutDescription
from randovania.layout.permalink import Permalink
from randovania.layout.preset_migration import VersionedPreset, InvalidPreset
from randovania.lib.status_update_lib import ProgressUpdateCallable
from randovania.resolver.exceptions import GenerationFailure


def persist_layout(history_dir: Path, description: LayoutDescription):
    history_dir.mkdir(parents=True, exist_ok=True)

    games = "-".join(sorted(game.short_name for game in description.all_games))

    date_format = datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S")
    file_path = history_dir.joinpath(
        f"{date_format}_{games}_{description.shareable_word_hash}.{description.file_extension()}")
    description.save_to_file(file_path)


class PresetMenu(QtWidgets.QMenu):
    action_customize: QtWidgets.QAction
    action_delete: QtWidgets.QAction
    action_history: QtWidgets.QAction
    action_export: QtWidgets.QAction
    action_duplicate: QtWidgets.QAction
    action_map_tracker: QtWidgets.QAction
    action_required_tricks: QtWidgets.QAction

    action_import: QtWidgets.QAction
    action_view_deleted: QtWidgets.QAction

    preset: Optional[VersionedPreset]

    def __init__(self, parent: QtWidgets.QWidget):
        super().__init__(parent)
        self.action_customize = QtWidgets.QAction(parent)
        self.action_delete = QtWidgets.QAction(parent)
        self.action_history = QtWidgets.QAction(parent)
        self.action_export = QtWidgets.QAction(parent)
        self.action_duplicate = QtWidgets.QAction(parent)
        self.action_map_tracker = QtWidgets.QAction(parent)
        self.action_required_tricks = QtWidgets.QAction(parent)
        self.action_import = QtWidgets.QAction(parent)
        self.action_view_deleted = QtWidgets.QAction(parent)

        self.action_customize.setText("Customize")
        self.action_delete.setText("Delete")
        self.action_history.setText("View previous versions")
        self.action_export.setText("Export")
        self.action_duplicate.setText("Duplicate")
        self.action_map_tracker.setText("Open map tracker")
        self.action_required_tricks.setText("View expected trick usage")
        self.action_import.setText("Import")
        self.action_view_deleted.setText("View deleted presets")

        self.addAction(self.action_customize)
        self.addAction(self.action_delete)
        self.addAction(self.action_history)
        self.addAction(self.action_export)
        self.addAction(self.action_duplicate)
        self.addAction(self.action_map_tracker)
        self.addAction(self.action_required_tricks)
        self.addSeparator()
        self.addAction(self.action_import)
        self.addAction(self.action_view_deleted)

        # TODO: Hide the ones that aren't implemented
        self.action_history.setVisible(False)
        self.action_view_deleted.setVisible(False)

    def set_preset(self, preset: Optional[VersionedPreset]):
        self.preset = preset

        for p in [self.action_delete, self.action_history, self.action_export]:
            p.setEnabled(preset is not None and preset.base_preset_uuid is not None)

        for p in [self.action_customize, self.action_duplicate, self.action_map_tracker, self.action_required_tricks]:
            p.setEnabled(preset is not None)


class GenerateSeedTab(QtWidgets.QWidget, BackgroundTaskMixin):
    _logic_settings_window: Optional[LogicSettingsWindow] = None
    _has_set_from_last_selected: bool = False
    _preset_menu: PresetMenu
    _action_delete: QtWidgets.QAction
    _original_show_event: Callable[[QtGui.QShowEvent], None]

    def __init__(self, window: Ui_MainWindow, window_manager: WindowManager, options: Options):
        super().__init__()

        self.window = window
        self._window_manager = window_manager
        self.failure_handler = GenerationFailureHandler(self)
        self._options = options

    def setup_ui(self):
        window = self.window
        window.create_preset_tree.window_manager = self._window_manager

        self._original_show_event = window.tab_create_seed.showEvent
        window.tab_create_seed.showEvent = self._tab_show_event

        # Progress
        self.background_tasks_button_lock_signal.connect(self.enable_buttons_with_background_tasks)
        self.progress_update_signal.connect(self.update_progress)
        self.window.stop_background_process_button.clicked.connect(self.stop_background_process)

        self.window.num_players_spin_box.setVisible(self._window_manager.is_preview_mode)
        self.window.create_generate_no_retry_button.setVisible(self._window_manager.is_preview_mode)

        # Menu
        self._preset_menu = PresetMenu(window)

        # Signals
        window.create_generate_button.clicked.connect(partial(self._generate_new_seed, True))
        window.create_generate_no_retry_button.clicked.connect(partial(self._generate_new_seed, True, retries=0))
        window.create_generate_race_button.clicked.connect(partial(self._generate_new_seed, False))
        window.create_preset_tree.itemSelectionChanged.connect(self._on_select_preset)
        window.create_preset_tree.customContextMenuRequested.connect(self._on_tree_context_menu)

        self._preset_menu.action_customize.triggered.connect(self._on_customize_preset)
        self._preset_menu.action_delete.triggered.connect(self._on_delete_preset)
        self._preset_menu.action_history.triggered.connect(self._on_view_preset_history)
        self._preset_menu.action_export.triggered.connect(self._on_export_preset)
        self._preset_menu.action_duplicate.triggered.connect(self._on_duplicate_preset)
        self._preset_menu.action_map_tracker.triggered.connect(self._on_open_map_tracker_for_preset)
        self._preset_menu.action_required_tricks.triggered.connect(self._on_open_required_tricks_for_preset)
        self._preset_menu.action_import.triggered.connect(self._on_import_preset)

        self._update_preset_tree_items()

    def _update_preset_tree_items(self):
        self.window.create_preset_tree.update_items()

    @asyncSlot()
    async def _do_migration(self):
        dialog = QtWidgets.QProgressDialog(
            ("Randovania changed where your presets are saved and a one-time migration is being performed.\n"
             "Further changes in old versions won't be migrated."),
            None,
            0, 1, self,
        )
        common_qt_lib.set_default_window_icon(dialog)
        dialog.setWindowTitle("Preset Migration")
        dialog.setAutoReset(False)
        dialog.setAutoClose(False)
        dialog.show()

        def on_update(current, target):
            dialog.setValue(current)
            dialog.setMaximum(target)

        await self._window_manager.preset_manager.migrate_from_old_path(on_update)
        self._update_preset_tree_items()
        dialog.setCancelButtonText("Ok")

    def _tab_show_event(self, event: QtGui.QShowEvent):
        if self._window_manager.preset_manager.should_do_migration():
            QTimer.singleShot(0, self._do_migration)

        return self._original_show_event(event)

    @property
    def _current_preset_data(self) -> Optional[VersionedPreset]:
        return self.window.create_preset_tree.current_preset_data

    def enable_buttons_with_background_tasks(self, value: bool):
        self.window.stop_background_process_button.setEnabled(not value)
        self.window.create_generate_button.setEnabled(value)
        self.window.create_generate_race_button.setEnabled(value)

    def _add_new_preset(self, preset: VersionedPreset):
        with self._options as options:
            options.selected_preset_uuid = preset.uuid

        self._window_manager.preset_manager.add_new_preset(preset)
        self._update_preset_tree_items()
        self.window.create_preset_tree.select_preset(preset)

    @asyncSlot()
    async def _on_customize_preset(self):
        if self._logic_settings_window is not None:
            self._logic_settings_window.raise_()
            return

        old_preset = self._current_preset_data.get_preset()
        if old_preset.base_preset_uuid is None:
            old_preset = old_preset.fork()
        editor = PresetEditor(old_preset)
        self._logic_settings_window = LogicSettingsWindow(self._window_manager, editor)

        self._logic_settings_window.on_preset_changed(editor.create_custom_preset_with())
        editor.on_changed = lambda: self._logic_settings_window.on_preset_changed(editor.create_custom_preset_with())

        result = await async_dialog.execute_dialog(self._logic_settings_window)
        self._logic_settings_window = None

        if result == QtWidgets.QDialog.Accepted:
            self._add_new_preset(VersionedPreset.with_preset(editor.create_custom_preset_with()))

    def _on_delete_preset(self):
        self._window_manager.preset_manager.delete_preset(self._current_preset_data)
        self._update_preset_tree_items()
        self._on_select_preset()

    def _on_view_preset_history(self):
        pass

    def _on_export_preset(self):
        path = common_qt_lib.prompt_user_for_preset_file(self._window_manager, new_file=True)
        if path is not None:
            self._current_preset_data.save_to_file(path)

    def _on_duplicate_preset(self):
        old_preset = self._current_preset_data
        self._add_new_preset(VersionedPreset.with_preset(old_preset.get_preset().fork()))

    def _on_open_map_tracker_for_preset(self):
        self._window_manager.open_map_tracker(self._current_preset_data.get_preset())

    def _on_open_required_tricks_for_preset(self):
        from randovania.gui.dialog.trick_usage_popup import TrickUsagePopup
        self._trick_usage_popup = TrickUsagePopup(self, self._window_manager, self._current_preset_data.get_preset())
        self._trick_usage_popup.setWindowModality(QtCore.Qt.WindowModal)
        self._trick_usage_popup.open()

    def _on_import_preset(self):
        path = common_qt_lib.prompt_user_for_preset_file(self._window_manager, new_file=False)
        if path is not None:
            self.import_preset_file(path)

    def import_preset_file(self, path: Path):
        preset = VersionedPreset.from_file_sync(path)
        try:
            preset.get_preset()
        except InvalidPreset:
            QtWidgets.QMessageBox.critical(
                self._window_manager,
                "Error loading preset",
                "The file at '{}' contains an invalid preset.".format(path)
            )
            return

        existing_preset = self._window_manager.preset_manager.preset_for_uuid(preset.uuid)
        if existing_preset is not None:
            user_response = QtWidgets.QMessageBox.warning(
                self._window_manager,
                "Preset ID conflict",
                "The new preset '{}' has the same ID as existing '{}'. Do you want to overwrite it?".format(
                    preset.name,
                    existing_preset.name,
                ),
                QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No | QtWidgets.QMessageBox.Cancel,
                QtWidgets.QMessageBox.Cancel
            )
            if user_response == QtWidgets.QMessageBox.Cancel:
                return
            elif user_response == QtWidgets.QMessageBox.No:
                preset = VersionedPreset.with_preset(dataclasses.replace(preset.get_preset(), uuid=uuid.uuid4()))

        self._add_new_preset(preset)

    def _on_select_preset(self):
        preset_data = self._current_preset_data
        self.on_preset_changed(preset_data)

        if preset_data is not None:
            with self._options as options:
                options.selected_preset_uuid = preset_data.uuid

    def _on_tree_context_menu(self, pos: QtCore.QPoint):
        item: QtWidgets.QTreeWidgetItem = self.window.create_preset_tree.itemAt(pos)
        preset = None
        if item is not None:
            preset = self.window.create_preset_tree.preset_for_item(item)

        self._preset_menu.set_preset(preset)
        self._preset_menu.exec_(QtGui.QCursor.pos())

    # Generate seed

    def _generate_new_seed(self, spoiler: bool, retries: Optional[int] = None):
        preset = self._current_preset_data
        num_players = self.window.num_players_spin_box.value()

        self.generate_seed_from_permalink(Permalink(
            seed_number=random.randint(0, 2 ** 31),
            spoiler=spoiler,
            presets={
                i: preset.get_preset()
                for i in range(num_players)
            },
        ), retries=retries)

    def generate_seed_from_permalink(self, permalink: Permalink, retries: Optional[int] = None):
        def work(progress_update: ProgressUpdateCallable):
            try:
                layout = simplified_patcher.generate_layout(progress_update=progress_update,
                                                            permalink=permalink,
                                                            options=self._options,
                                                            retries=retries)
                progress_update(f"Success! (Seed hash: {layout.shareable_hash})", 1)
                persist_layout(self._options.game_history_path, layout)
                self._window_manager.open_game_details(layout)

            except GenerationFailure as generate_exception:
                self.failure_handler.handle_failure(generate_exception)
                progress_update("Generation Failure: {}".format(generate_exception), -1)

        if self._window_manager.is_preview_mode:
            logging.info(f"Permalink: {permalink.as_base64_str}")
        self.run_in_background_thread(work, "Creating a seed...")

    def on_options_changed(self, options: Options):
        self.window.create_preset_tree.set_show_experimental(options.experimental_games)

        if not self._has_set_from_last_selected:
            self._has_set_from_last_selected = True
            preset = self._window_manager.preset_manager.preset_for_uuid(options.selected_preset_uuid)
            if preset is None:
                preset = self._window_manager.preset_manager.default_preset
            self.window.create_preset_tree.select_preset(preset)

    def on_preset_changed(self, preset: Optional[VersionedPreset]):
        can_generate = False
        if preset is None:
            description = "Please select a preset from the list, not a game."

        else:
            try:
                raw_preset = preset.get_preset()
                can_generate = True
                description = f"<p style='font-weight:600;'>{raw_preset.name}</p><p>{raw_preset.description}</p>"
                description += preset_describer.merge_categories(preset_describer.describe(raw_preset))

            except InvalidPreset as e:
                logging.exception(f"Invalid preset for {preset.name}")
                description = (
                    f"Preset {preset.name} can't be used as it contains the following error:"
                    f"\n{e.original_exception}\n"
                    f"\nPlease open edit the preset file with id {preset.uuid} manually or delete this preset."
                )

        self.window.create_preset_description.setText(description)
        for btn in [self.window.create_generate_button, self.window.create_generate_race_button]:
            btn.setEnabled(can_generate)

    def update_progress(self, message: str, percentage: int):
        self.window.progress_label.setText(message)
        if "Aborted" in message:
            percentage = 0
        if percentage >= 0:
            self.window.progress_bar.setRange(0, 100)
            self.window.progress_bar.setValue(percentage)
        else:
            self.window.progress_bar.setRange(0, 0)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'widget_pickup_style.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *

from randovania.gui.lib.scroll_protected import ScrollProtectedComboBox


class Ui_PickupStyleWidget(object):
    def setupUi(self, PickupStyleWidget):
        if not PickupStyleWidget.objectName():
            PickupStyleWidget.setObjectName(u"PickupStyleWidget")
        PickupStyleWidget.resize(534, 116)
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(PickupStyleWidget.sizePolicy().hasHeightForWidth())
        PickupStyleWidget.setSizePolicy(sizePolicy)
        PickupStyleWidget.setMaximumSize(QSize(16777215, 16777215))
        self.verticalLayout = QVBoxLayout(PickupStyleWidget)
        self.verticalLayout.setSpacing(6)
        self.verticalLayout.setContentsMargins(11, 11, 11, 11)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.pickup_style_box = QGroupBox(PickupStyleWidget)
        self.pickup_style_box.setObjectName(u"pickup_style_box")
        sizePolicy.setHeightForWidth(self.pickup_style_box.sizePolicy().hasHeightForWidth())
        self.pickup_style_box.setSizePolicy(sizePolicy)
        self.verticalLayout_2 = QVBoxLayout(self.pickup_style_box)
        self.verticalLayout_2.setSpacing(6)
        self.verticalLayout_2.setContentsMargins(11, 11, 11, 11)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.pickup_model_combo = ScrollProtectedComboBox(self.pickup_style_box)
        self.pickup_model_combo.addItem("")
        self.pickup_model_combo.addItem("")
        self.pickup_model_combo.addItem("")
        self.pickup_model_combo.addItem("")
        self.pickup_model_combo.setObjectName(u"pickup_model_combo")

        self.verticalLayout_2.addWidget(self.pickup_model_combo)

        self.pickup_data_source_label = QLabel(self.pickup_style_box)
        self.pickup_data_source_label.setObjectName(u"pickup_data_source_label")

        self.verticalLayout_2.addWidget(self.pickup_data_source_label)

        self.pickup_data_source_combo = ScrollProtectedComboBox(self.pickup_style_box)
        self.pickup_data_source_combo.addItem("")
        self.pickup_data_source_combo.addItem("")
        self.pickup_data_source_combo.addItem("")
        self.pickup_data_source_combo.setObjectName(u"pickup_data_source_combo")

        self.verticalLayout_2.addWidget(self.pickup_data_source_combo)


        self.verticalLayout.addWidget(self.pickup_style_box)


        self.retranslateUi(PickupStyleWidget)

        QMetaObject.connectSlotsByName(PickupStyleWidget)
    # setupUi

    def retranslateUi(self, PickupStyleWidget):
        PickupStyleWidget.setWindowTitle(QCoreApplication.translate("PickupStyleWidget", u"Item Configuration", None))
        self.pickup_style_box.setTitle(QCoreApplication.translate("PickupStyleWidget", u"Pickup style", None))
        self.pickup_model_combo.setItemText(0, QCoreApplication.translate("PickupStyleWidget", u"Use correct item model, scan and name", None))
        self.pickup_model_combo.setItemText(1, QCoreApplication.translate("PickupStyleWidget", u"Use correct scan and name, hide the model", None))
        self.pickup_model_combo.setItemText(2, QCoreApplication.translate("PickupStyleWidget", u"Use correct name, hide the model and scan", None))
        self.pickup_model_combo.setItemText(3, QCoreApplication.translate("PickupStyleWidget", u"Hide the model, scan and name", None))

        self.pickup_data_source_label.setText(QCoreApplication.translate("PickupStyleWidget", u"When hiding some part of the pickup, it's replaced with:", None))
        self.pickup_data_source_combo.setItemText(0, QCoreApplication.translate("PickupStyleWidget", u"Energy Transfer Module/Nothing's data", None))
        self.pickup_data_source_combo.setItemText(1, QCoreApplication.translate("PickupStyleWidget", u"A random item data", None))
        self.pickup_data_source_combo.setItemText(2, QCoreApplication.translate("PickupStyleWidget", u"The data of the pickup in that place", None))

    # retranslateUi




import functools
import json
from pathlib import Path

from randovania import get_data_path
from randovania.game_description import data_reader
from randovania.game_description.data_reader import read_resource_database
from randovania.game_description.game_description import GameDescription
from randovania.game_description.item import item_database
from randovania.game_description.resources.resource_database import ResourceDatabase
from randovania.games import default_data
from randovania.games.game import RandovaniaGame


@functools.lru_cache()
def resource_database_for(game: RandovaniaGame) -> ResourceDatabase:
    return read_resource_database(game, default_data.read_json_then_binary(game)[1]["resource_database"])


@functools.lru_cache()
def game_description_for(game: RandovaniaGame) -> GameDescription:
    result = data_reader.decode_data(default_data.read_json_then_binary(game)[1])
    if result.game != game:
        raise ValueError(f"Game Description for {game} has game field {result.game}")
    return result


def _read_item_database_in_path(path: Path, game: RandovaniaGame) -> item_database.ItemDatabase:
    with path.joinpath("item-database.json").open() as database_file:
        item_database_data = json.load(database_file)

    return item_database.read_database(item_database_data, game)


@functools.lru_cache()
def item_database_for_game(game: RandovaniaGame):
    return _read_item_database_in_path(get_data_path().joinpath("item_database", game.value),
                                       game)


@functools.lru_cache()
def default_prime2_memo_data() -> dict:
    with get_data_path().joinpath("item_database", "prime2", "memo_data.json").open("r") as memo_data_file:
        memo_data = json.load(memo_data_file)

    TEMPLE_KEYS = ["Dark Agon Key", "Dark Torvus Key", "Ing Hive Key"]

    for i in range(1, 4):
        for temple_key in TEMPLE_KEYS:
            memo_data["{} {}".format(temple_key, i)] = memo_data[temple_key]

    for temple_key in TEMPLE_KEYS:
        memo_data.pop(temple_key)

    for i in range(1, 10):
        memo_data["Sky Temple Key {}".format(i)] = memo_data["Sky Temple Key"]
    memo_data.pop("Sky Temple Key")

    return memo_data



import dataclasses
from enum import Enum

from randovania.games.game import RandovaniaGame
from randovania.layout.base.cosmetic_patches import BaseCosmeticPatches


class CorruptionSuit(Enum):
    VARIA = 0
    GRAPPLE = 1
    PED = 2
    CORRUPTED_25 = 3
    CORRUPTED_50 = 4


@dataclasses.dataclass(frozen=True)
class CorruptionCosmeticPatches(BaseCosmeticPatches):
    random_door_colors: bool = False
    random_welding_colors: bool = False
    player_suit: CorruptionSuit = CorruptionSuit.VARIA

    @classmethod
    def default(cls) -> "CorruptionCosmeticPatches":
        return cls()

    @classmethod
    def game(cls):
        return RandovaniaGame.METROID_PRIME_CORRUPTION



from unittest.mock import MagicMock, call

import pytest

from randovania.game_connection.executor.dolphin_executor import DolphinExecutor
from randovania.game_connection.executor.memory_operation import MemoryOperation


@pytest.fixture(name="executor")
def dolphin_executor():
    executor = DolphinExecutor()
    executor.dolphin = MagicMock()
    return executor


@pytest.mark.asyncio
async def test_perform_memory_operations(executor: DolphinExecutor):
    executor.dolphin.follow_pointers.return_value = 0x80003000
    executor.dolphin.read_bytes.side_effect = [b"A" * 50, b"B" * 30, b"C" * 10]

    # Run
    result = await executor.perform_memory_operations([
        MemoryOperation(0x80001000, offset=20, read_byte_count=50),
        MemoryOperation(0x80001000, offset=10, read_byte_count=30, write_bytes=b"1" * 30),
        MemoryOperation(0x80002000, read_byte_count=10),
    ])

    # Assert
    assert list(result.values()) == [b"A" * 50, b"B" * 30, b"C" * 10]
    executor.dolphin.follow_pointers.assert_called_once_with(0x80001000, [0x0])
    executor.dolphin.read_bytes.assert_has_calls([
        call(0x80003000 + 20, 50),
        call(0x80003000 + 10, 30),
        call(0x80002000, 10),
    ])
    executor.dolphin.write_bytes.assert_called_once_with(0x80003000 + 10, b"1" * 30)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'connections_editor.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_ConnectionEditor(object):
    def setupUi(self, ConnectionEditor):
        if not ConnectionEditor.objectName():
            ConnectionEditor.setObjectName(u"ConnectionEditor")
        ConnectionEditor.resize(794, 393)
        ConnectionEditor.setMinimumSize(QSize(600, 0))
        self.gridLayout_2 = QGridLayout(ConnectionEditor)
        self.gridLayout_2.setObjectName(u"gridLayout_2")
        self.visualizer_scroll = QScrollArea(ConnectionEditor)
        self.visualizer_scroll.setObjectName(u"visualizer_scroll")
        self.visualizer_scroll.setWidgetResizable(True)
        self.visualizer_contents = QWidget()
        self.visualizer_contents.setObjectName(u"visualizer_contents")
        self.visualizer_contents.setGeometry(QRect(0, 0, 774, 345))
        self.contents_layout = QVBoxLayout(self.visualizer_contents)
        self.contents_layout.setObjectName(u"contents_layout")
        self.visualizer_scroll.setWidget(self.visualizer_contents)

        self.gridLayout_2.addWidget(self.visualizer_scroll, 0, 0, 1, 1)

        self.button_box = QDialogButtonBox(ConnectionEditor)
        self.button_box.setObjectName(u"button_box")
        self.button_box.setOrientation(Qt.Horizontal)
        self.button_box.setStandardButtons(QDialogButtonBox.Cancel|QDialogButtonBox.Ok)

        self.gridLayout_2.addWidget(self.button_box, 1, 0, 1, 1)


        self.retranslateUi(ConnectionEditor)
        self.button_box.accepted.connect(ConnectionEditor.accept)
        self.button_box.rejected.connect(ConnectionEditor.reject)

        QMetaObject.connectSlotsByName(ConnectionEditor)
    # setupUi

    def retranslateUi(self, ConnectionEditor):
        ConnectionEditor.setWindowTitle(QCoreApplication.translate("ConnectionEditor", u"Connection Editor", None))
    # retranslateUi




import logging
import traceback
from pathlib import Path
from typing import Optional

from PySide2 import QtWidgets
from PySide2.QtCore import Signal

from randovania.games.patcher import Patcher
from randovania.games.patchers.exceptions import ExportFailure
from randovania.gui.dialog.game_input_dialog import GameInputDialog
from randovania.gui.lib import common_qt_lib, async_dialog
from randovania.gui.lib.background_task_mixin import BackgroundTaskMixin
from randovania.layout.layout_description import LayoutDescription
from randovania.lib.status_update_lib import ProgressUpdateCallable


async def export_game(
        patcher: Patcher,
        input_dialog: GameInputDialog,
        patch_data: dict,
        internal_copies_path: Path,
        layout_for_spoiler: Optional[LayoutDescription],
        background: BackgroundTaskMixin,
        progress_update_signal: Signal(str, int),
):
    input_file = input_dialog.input_file
    output_file = input_dialog.output_file
    auto_save_spoiler = input_dialog.auto_save_spoiler

    if patcher.is_busy:
        return await async_dialog.message_box(
            None, QtWidgets.QMessageBox.Critical,
            "Can't export game",
            "Error: Unable to export multiple games at the same time and "
            "another window is exporting a game right now.")

    def work(progress_update: ProgressUpdateCallable):
        patcher.patch_game(input_file, output_file, patch_data, internal_copies_path,
                           progress_update=progress_update)

        if auto_save_spoiler and layout_for_spoiler is not None and layout_for_spoiler.permalink.spoiler:
            layout_for_spoiler.save_to_file(output_file.with_suffix(f".{LayoutDescription.file_extension()}"))

        progress_update(f"Finished!", 1)

    try:
        await background.run_in_background_async(work, "Exporting...")
    except Exception as e:
        logging.exception("Unable to export game")

        message = str(e)
        detailed = None
        if isinstance(e, ExportFailure):
            detailed = e.detailed_text()
        elif e.__traceback__ is not None:
            detailed = "".join(traceback.format_tb(e.__traceback__))

        progress_update_signal.emit(f"Unable to export game: {message}", None)
        box = QtWidgets.QMessageBox(
            QtWidgets.QMessageBox.Critical,
            "Unable to export game",
            message + ".\nPress show 'Show Details' for more information.",
            QtWidgets.QMessageBox.Ok,
        )
        common_qt_lib.set_default_window_icon(box)
        if detailed is not None:
            box.setDetailedText(detailed)
        await async_dialog.execute_dialog(box)



from typing import Iterator

from randovania.game_description.resources.pickup_entry import PickupEntry
from randovania.game_description.resources.resource_database import ResourceDatabase
from randovania.generator.item_pool.pickup_creator import create_ammo_expansion
from randovania.layout.base.ammo_configuration import AmmoConfiguration


def add_ammo(resource_database: ResourceDatabase,
             ammo_configuration: AmmoConfiguration,
             ) -> Iterator[PickupEntry]:
    """
    Creates the necessary pickups for the given ammo_configuration.
    :param resource_database:
    :param ammo_configuration:
    :return:
    """
    for ammo, state in ammo_configuration.items_state.items():
        for _ in range(state.pickup_count):
            yield create_ammo_expansion(ammo, state.ammo_count, state.requires_major_item, resource_database)



import dataclasses
from typing import List, Tuple, Set, Optional

from randovania.game_connection.connection_base import Inventory
from randovania.game_connection.executor.memory_operation import MemoryOperationExecutor, MemoryOperation
from randovania.game_description.resources.pickup_entry import PickupEntry
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.game_description.world.world import World
from randovania.games.game import RandovaniaGame


@dataclasses.dataclass(frozen=True)
class RemotePatch:
    """Format of the contents depends exclusively on each implementation."""
    memory_operations: List[MemoryOperation]


class RemoteConnector:
    @property
    def game_enum(self) -> RandovaniaGame:
        raise NotImplementedError()

    async def is_this_version(self, executor: MemoryOperationExecutor) -> bool:
        """Returns True if the accessible memory matches the version of this connector."""
        raise NotImplementedError()

    async def current_game_status(self, executor: MemoryOperationExecutor) -> Tuple[bool, Optional[World]]:
        """
        Fetches the world the player's currently at, or None if they're not in-game.
        :param executor:
        :return: bool indicating if there's a pending `execute_remote_patches` operation.
        """
        raise NotImplementedError()

    async def get_inventory(self, executor: MemoryOperationExecutor) -> Inventory:
        """Fetches the inventory represented by the given game memory."""
        raise NotImplementedError()

    async def known_collected_locations(self, executor: MemoryOperationExecutor,
                                        ) -> Tuple[Set[PickupIndex], List[RemotePatch]]:
        """Fetches pickup indices that have been collected.
        The list may return less than all collected locations, depending on implementation details.
        This function also returns a list of remote patches that must be performed via `execute_remote_patches`.
        """
        raise NotImplementedError()

    async def find_missing_remote_pickups(self, executor: MemoryOperationExecutor, inventory: Inventory,
                                          remote_pickups: Tuple[Tuple[str, PickupEntry], ...],
                                          in_cooldown: bool,
                                          ) -> Tuple[List[RemotePatch], bool]:
        """
        Determines if any of the remote_pickups needs to be written to executor.
        :param executor:
        :param inventory: The player's inventory, as given by `get_inventory`.
        :param remote_pickups: Ordered list of pickups sent from other players, with the name of the player.
        :param in_cooldown: If sending new pickups is on cooldown.
        :return: List of patches to give one missing pickup. A bool indicating that a message will be displayed.
        """
        raise NotImplementedError()

    async def execute_remote_patches(self, executor: MemoryOperationExecutor, patches: List[RemotePatch]) -> None:
        """
        Executes a given set of patches on the given memory operator. Should only be called if the bool returned by
        `current_game_status` is False, but validation of this fact is implementation-dependant.
        :param executor:
        :param patches: List of patches to execute
        :return:
        """
        raise NotImplementedError()



from typing import Optional

import pytest
from mock import MagicMock, AsyncMock, call, ANY

import randovania
from randovania.games.game import RandovaniaGame
from randovania.server.discord import preset_lookup


@pytest.mark.asyncio
async def test_on_message_from_bot(mocker):
    mock_look_for: AsyncMock = mocker.patch("randovania.server.discord.preset_lookup.look_for_permalinks",
                                            new_callable=AsyncMock)
    client = MagicMock()
    cog = preset_lookup.PermalinkLookupCog(None, client)

    message = MagicMock()
    message.author = client.user

    # Run
    await cog.on_message(message)

    # Assert
    mock_look_for.assert_not_awaited()


@pytest.mark.parametrize(["guild", "expected"], [
    (None, True),
    (0, False),
    (1234, True),
])
@pytest.mark.asyncio
async def test_on_message_wrong_place(mocker, guild: Optional[int], expected):
    mock_look_for: AsyncMock = mocker.patch("randovania.server.discord.preset_lookup.look_for_permalinks",
                                            new_callable=AsyncMock)
    cog = preset_lookup.PermalinkLookupCog({"guild": 1234}, MagicMock())

    message = MagicMock()
    message.channel.name = "the_expected_name"
    if guild is None:
        message.guild = None
    else:
        message.guild.id = guild

    # Run
    await cog.on_message(message)

    # Assert
    if expected:
        mock_look_for.assert_awaited_once_with(message)
    else:
        mock_look_for.assert_not_awaited()


@pytest.mark.parametrize("is_solo", [False, True])
@pytest.mark.parametrize("has_multiple", [False, True])
@pytest.mark.asyncio
async def test_look_for_permalinks(mocker, is_solo, has_multiple):
    preset = MagicMock()
    preset.game = RandovaniaGame.METROID_PRIME_ECHOES
    permalink_1 = MagicMock()
    permalink_1.player_count = 1 if is_solo else 2
    permalink_1.get_preset.return_value = preset

    permalink_2 = MagicMock()
    embed = MagicMock()

    mock_embed: MagicMock = mocker.patch("discord.Embed", side_effect=[embed])
    mock_create_actionrow = mocker.patch("discord_slash.utils.manage_components.create_actionrow")

    mock_describe: MagicMock = mocker.patch("randovania.gui.lib.preset_describer.describe",
                                            return_value=[
                                                ("General", ["Foo", "Bar"]),
                                                ("Other", ["X", "Y"]),
                                            ])
    mock_from_str: MagicMock = mocker.patch(
        "randovania.layout.permalink.Permalink.from_str",
        side_effect=[permalink_1, permalink_2] if has_multiple else [permalink_1],
    )

    message = AsyncMock()
    if has_multiple:
        message.content = "yu4abbceWfLI- `yua73123yWdLI-` foo???"
    else:
        message.content = "yu4abbceWfLI- foo???"

    # Run
    await preset_lookup.look_for_permalinks(message)

    # Assert
    if has_multiple:
        mock_from_str.assert_has_calls([
            call("yu4abbceWfLI-"),
            call("yua73123yWdLI-"),
        ])
    else:
        mock_from_str.assert_called_once_with("yu4abbceWfLI-")

    if is_solo:
        permalink_1.get_preset.assert_called_once_with(0)
        mock_describe.assert_called_once_with(preset)
    else:
        permalink_1.get_preset.assert_not_called()
        mock_describe.assert_not_called()

    mock_embed.assert_called_once_with(
        title="`yu4abbceWfLI-`", description=f"{permalink_1.player_count} player multiworld permalink",
    )
    if is_solo:
        assert embed.description == "Metroid Prime 2: Echoes permalink for Randovania {}".format(randovania.VERSION)
        embed.add_field.assert_has_calls([
            call(name="General", value="Foo\nBar", inline=True),
            call(name="Other", value="X\nY", inline=True),
        ])

    content = None
    if has_multiple:
        content = "Multiple permalinks found, using only the first."
    message.reply.assert_awaited_once_with(
        content=content,
        embed=embed,
        components=[mock_create_actionrow.return_value],
        mention_author=False,
    )


@pytest.mark.asyncio
async def test_reply_for_preset(mocker):
    mock_describe: MagicMock = mocker.patch("randovania.gui.lib.preset_describer.describe",
                                            return_value=[
                                                ("General", ["Foo", "Bar"]),
                                                ("Other", ["X", "Y"]),
                                            ])
    message = AsyncMock()
    versioned_preset = MagicMock()
    preset = versioned_preset.get_preset.return_value
    embed = MagicMock()

    mock_embed: MagicMock = mocker.patch("discord.Embed", side_effect=[embed])

    # Run
    await preset_lookup.reply_for_preset(message, versioned_preset)

    # Assert
    mock_embed.assert_called_once_with(title=preset.name, description=preset.description)
    embed.add_field.assert_has_calls([
        call(name="General", value="Foo\nBar", inline=True),
        call(name="Other", value="X\nY", inline=True),
    ])
    message.reply.assert_awaited_once_with(embed=embed, mention_author=False)
    mock_describe.assert_called_once_with(preset)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'logic_settings_window.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_LogicSettingsWindow(object):
    def setupUi(self, LogicSettingsWindow):
        if not LogicSettingsWindow.objectName():
            LogicSettingsWindow.setObjectName(u"LogicSettingsWindow")
        LogicSettingsWindow.resize(700, 768)
        self.verticalLayout = QVBoxLayout(LogicSettingsWindow)
        self.verticalLayout.setSpacing(6)
        self.verticalLayout.setContentsMargins(11, 11, 11, 11)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.name_layout = QHBoxLayout()
        self.name_layout.setSpacing(6)
        self.name_layout.setObjectName(u"name_layout")
        self.name_label = QLabel(LogicSettingsWindow)
        self.name_label.setObjectName(u"name_label")

        self.name_layout.addWidget(self.name_label)

        self.name_edit = QLineEdit(LogicSettingsWindow)
        self.name_edit.setObjectName(u"name_edit")

        self.name_layout.addWidget(self.name_edit)


        self.verticalLayout.addLayout(self.name_layout)

        self.main_tab_widget = QTabWidget(LogicSettingsWindow)
        self.main_tab_widget.setObjectName(u"main_tab_widget")
        self.main_tab_widget.setTabPosition(QTabWidget.North)
        self.main_tab_widget.setTabShape(QTabWidget.Rounded)
        self.main_tab_widget.setTabBarAutoHide(False)
        self.logic_tab = QWidget()
        self.logic_tab.setObjectName(u"logic_tab")
        font = QFont()
        font.setBold(False)
        font.setWeight(50)
        self.logic_tab.setFont(font)
        self.logic_tab_layout = QVBoxLayout(self.logic_tab)
        self.logic_tab_layout.setSpacing(6)
        self.logic_tab_layout.setContentsMargins(11, 11, 11, 11)
        self.logic_tab_layout.setObjectName(u"logic_tab_layout")
        self.logic_tab_layout.setContentsMargins(2, 2, 2, 2)
        self.logic_tab_widget = QTabWidget(self.logic_tab)
        self.logic_tab_widget.setObjectName(u"logic_tab_widget")

        self.logic_tab_layout.addWidget(self.logic_tab_widget)

        self.main_tab_widget.addTab(self.logic_tab, "")
        self.patches_tab = QWidget()
        self.patches_tab.setObjectName(u"patches_tab")
        self.patches_tab_layout = QVBoxLayout(self.patches_tab)
        self.patches_tab_layout.setSpacing(6)
        self.patches_tab_layout.setContentsMargins(11, 11, 11, 11)
        self.patches_tab_layout.setObjectName(u"patches_tab_layout")
        self.patches_tab_layout.setContentsMargins(2, 2, 2, 2)
        self.patches_tab_widget = QTabWidget(self.patches_tab)
        self.patches_tab_widget.setObjectName(u"patches_tab_widget")

        self.patches_tab_layout.addWidget(self.patches_tab_widget)

        self.main_tab_widget.addTab(self.patches_tab, "")

        self.verticalLayout.addWidget(self.main_tab_widget)

        self.button_box = QDialogButtonBox(LogicSettingsWindow)
        self.button_box.setObjectName(u"button_box")
        self.button_box.setStandardButtons(QDialogButtonBox.Cancel|QDialogButtonBox.Save)

        self.verticalLayout.addWidget(self.button_box)


        self.retranslateUi(LogicSettingsWindow)

        self.main_tab_widget.setCurrentIndex(0)
        self.logic_tab_widget.setCurrentIndex(-1)
        self.patches_tab_widget.setCurrentIndex(-1)


        QMetaObject.connectSlotsByName(LogicSettingsWindow)
    # setupUi

    def retranslateUi(self, LogicSettingsWindow):
        LogicSettingsWindow.setWindowTitle(QCoreApplication.translate("LogicSettingsWindow", u"Customize Preset", None))
        self.name_label.setText(QCoreApplication.translate("LogicSettingsWindow", u"Name:", None))
        self.main_tab_widget.setTabText(self.main_tab_widget.indexOf(self.logic_tab), QCoreApplication.translate("LogicSettingsWindow", u"Randomizer Logic", None))
        self.main_tab_widget.setTabText(self.main_tab_widget.indexOf(self.patches_tab), QCoreApplication.translate("LogicSettingsWindow", u"Game Modifications", None))
    # retranslateUi




import pytest

from randovania.bitpacking import bitpacking
from randovania.bitpacking.bitpacking import BitPackDecoder
from randovania.game_description.world.area_location import AreaLocation
from randovania.game_description.world.teleporter import Teleporter
from randovania.games.game import RandovaniaGame
from randovania.layout.lib.teleporters import TeleporterConfiguration, TeleporterShuffleMode, TeleporterList, \
    TeleporterTargetList


def _m(encoded: bytes, bit_count: int, mode: str = "vanilla", skip_final_bosses=False,
       allow_unvisited_room_names=True, excluded_teleporters=None, excluded_targets=None):
    if excluded_teleporters is None:
        excluded_teleporters = []
    if excluded_targets is None:
        excluded_targets = []
    return {
        "encoded": encoded,
        "bit_count": bit_count,
        "json": {
            "mode": mode,
            "skip_final_bosses": skip_final_bosses,
            "allow_unvisited_room_names": allow_unvisited_room_names,
            "excluded_teleporters": excluded_teleporters,
            "excluded_targets": excluded_targets,
        },
    }


def _a(world, area, instance_id=None):
    if instance_id is not None:
        return Teleporter(world, area, instance_id).as_json
    return AreaLocation(world, area).as_json


@pytest.fixture(
    params=[
        _m(b'\x08', 5),
        _m(b'\x18', 5, skip_final_bosses=True),
        _m(b'\xb1', 8, mode="one-way-elevator"),
        _m(b'\xb81d', 22, mode="one-way-elevator", excluded_teleporters=[_a(1119434212, 1473133138, 122)]),
    ],
    name="with_data")
def _with_data(request):
    game = RandovaniaGame.METROID_PRIME_ECHOES
    reference = TeleporterConfiguration(
        mode=TeleporterShuffleMode.VANILLA,
        skip_final_bosses=False,
        allow_unvisited_room_names=False,
        excluded_teleporters=TeleporterList(frozenset(), game),
        excluded_targets=TeleporterTargetList(frozenset(), game),
    )
    return (reference, request.param["encoded"], request.param["bit_count"],
            TeleporterConfiguration.from_json(request.param["json"], game=game))


def test_decode(with_data):
    # Setup
    reference, data, _, expected = with_data

    # Run
    decoder = BitPackDecoder(data)
    result = TeleporterConfiguration.bit_pack_unpack(decoder, {"reference": reference})

    # Assert
    assert result == expected


def test_encode(with_data):
    # Setup
    reference, expected_bytes, expected_bit_count, value = with_data

    # Run
    result, bit_count = bitpacking.pack_results_and_bit_count(value.bit_pack_encode({"reference": reference}))

    # Assert
    assert result == expected_bytes
    assert bit_count == expected_bit_count



import pytest

from randovania.bitpacking import bitpacking
from randovania.bitpacking.bitpacking import BitPackDecoder
from randovania.game_description.item.major_item import MajorItem
from randovania.games.game import RandovaniaGame
from randovania.layout.base.major_item_state import MajorItemState


@pytest.fixture(
    params=[
        {"encoded": b'\x00', "bit_count": 3, "json": {}},
        {"encoded": b'@', "bit_count": 3, "json": {"num_shuffled_pickups": 1}},
        {"encoded": b'\x80', "bit_count": 7, "json": {"num_shuffled_pickups": 2}},
        {"encoded": b'\x84', "bit_count": 7, "json": {"num_shuffled_pickups": 3}},
        {"encoded": b'\xa2\xc8', "bit_count": 14, "json": {"num_shuffled_pickups": 99}},

        # Energy Tank
        {"encoded": b'\x92\x80', "bit_count": 12, "progression": 42, "json": {"num_shuffled_pickups": 6,
                                                                        "num_included_in_starting_items": 10}},

        # Ammo
        {"encoded": b'\x17', "bit_count": 8, "ammo_index": (43,), "json": {"included_ammo": [7]}},
        {"encoded": b'\x00', "bit_count": 4, "ammo_index": (45,), "json": {"included_ammo": [0]}},
        {"encoded": b'\x10P', "bit_count": 12, "ammo_index": (45,), "json": {"included_ammo": [5]}},
        {"encoded": b'\x00', "bit_count": 4, "ammo_index": (45, 46), "json": {"included_ammo": [0, 0]}},
        {"encoded": b'\x1c\xb0', "bit_count": 13, "ammo_index": (45, 46), "json": {"included_ammo": [150, 150]}},
        {"encoded": b'\x176@', "bit_count": 21, "ammo_index": (45, 46), "json": {"included_ammo": [230, 200]}},
    ],
    name="major_item_state")
def _major_item_state(request, echoes_item_database, generic_item_category):
    encoded: bytes = request.param["encoded"]

    item = MajorItem(
        game=RandovaniaGame.METROID_PRIME_ECHOES,
        name="Item Name",
        item_category=generic_item_category,
        broad_category=generic_item_category,
        model_name="Model Name",
        progression=(request.param.get("progression", 0),),
        ammo_index=request.param.get("ammo_index", ()),
        required=True,
        original_index=None,
        probability_offset=0,
    )
    included_ammo = tuple(0 for _ in request.param["json"].get("included_ammo", []))
    reference = MajorItemState(included_ammo=included_ammo)
    return item, encoded, request.param["bit_count"], MajorItemState.from_json(request.param["json"]), reference


def test_decode(major_item_state):
    # Setup
    item, data, _, expected, reference = major_item_state

    # Run
    decoder = BitPackDecoder(data)
    result = MajorItemState.bit_pack_unpack(decoder, item, reference=reference)

    # Assert
    assert result == expected


def test_encode(major_item_state):
    # Setup
    item, expected_bytes, expected_bit_count, value, reference = major_item_state

    # Run
    result, bit_count = bitpacking.pack_results_and_bit_count(value.bit_pack_encode(item, reference=reference))

    # Assert
    assert bit_count == expected_bit_count
    assert result == expected_bytes


def test_blank_as_json():
    assert MajorItemState().as_json == {}


def test_blank_from_json():
    assert MajorItemState.from_json({}) == MajorItemState()



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'background_process_dialog.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_BackgroundProcessDialog(object):
    def setupUi(self, BackgroundProcessDialog):
        if not BackgroundProcessDialog.objectName():
            BackgroundProcessDialog.setObjectName(u"BackgroundProcessDialog")
        BackgroundProcessDialog.resize(449, 80)
        self.gridLayout = QGridLayout(BackgroundProcessDialog)
        self.gridLayout.setSpacing(6)
        self.gridLayout.setContentsMargins(11, 11, 11, 11)
        self.gridLayout.setObjectName(u"gridLayout")
        self.update_button = QPushButton(BackgroundProcessDialog)
        self.update_button.setObjectName(u"update_button")

        self.gridLayout.addWidget(self.update_button, 1, 1, 1, 1)

        self.progress_bar = QProgressBar(BackgroundProcessDialog)
        self.progress_bar.setObjectName(u"progress_bar")
        self.progress_bar.setValue(0)
        self.progress_bar.setTextVisible(True)
        self.progress_bar.setInvertedAppearance(False)

        self.gridLayout.addWidget(self.progress_bar, 1, 0, 1, 1)

        self.progress_label = QLabel(BackgroundProcessDialog)
        self.progress_label.setObjectName(u"progress_label")
        sizePolicy = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.progress_label.sizePolicy().hasHeightForWidth())
        self.progress_label.setSizePolicy(sizePolicy)
        font = QFont()
        font.setPointSize(7)
        self.progress_label.setFont(font)
        self.progress_label.setAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignVCenter)
        self.progress_label.setWordWrap(True)

        self.gridLayout.addWidget(self.progress_label, 0, 0, 1, 3)


        self.retranslateUi(BackgroundProcessDialog)

        QMetaObject.connectSlotsByName(BackgroundProcessDialog)
    # setupUi

    def retranslateUi(self, BackgroundProcessDialog):
        BackgroundProcessDialog.setWindowTitle(QCoreApplication.translate("BackgroundProcessDialog", u"Import permalink", None))
        self.update_button.setText(QCoreApplication.translate("BackgroundProcessDialog", u"Stop", None))
        self.progress_label.setText(QCoreApplication.translate("BackgroundProcessDialog", u"<empty>", None))
    # retranslateUi




from typing import Optional, Tuple

from randovania.game_description.item.ammo import Ammo
from randovania.game_description.item.item_category import GENERIC_KEY_CATEGORY, USELESS_ITEM_CATEGORY, ItemCategory
from randovania.game_description.item.major_item import MajorItem
from randovania.game_description.resources.pickup_entry import PickupEntry, PickupModel
from randovania.game_description.resources.resource_database import ResourceDatabase
from randovania.game_description.resources.resource_info import ResourceQuantity
from randovania.games.game import RandovaniaGame
from randovania.games.prime import corruption_items, echoes_items
from randovania.games.prime import prime_items
from randovania.layout.base.major_item_state import MajorItemState


def create_major_item(item: MajorItem,
                      state: MajorItemState,
                      include_percentage: bool,
                      resource_database: ResourceDatabase,
                      ammo: Optional[Ammo],
                      ammo_requires_major_item: bool,

                      ) -> PickupEntry:
    """
    Creates a Pickup for the given MajorItem
    :param include_percentage:
    :param state:
    :param item:
    :param resource_database:
    :param ammo:
    :param ammo_requires_major_item:
    :return:
    """

    extra_resources = [
        (resource_database.get_item(ammo_index), ammo_count)
        for ammo_index, ammo_count in zip(item.ammo_index, state.included_ammo)
    ]
    if include_percentage and resource_database.item_percentage is not None:
        extra_resources.append((resource_database.item_percentage, 1))

    def _create_resources(base_resource: Optional[int]) -> ResourceQuantity:
        # FIXME: hacky quantity for Hazard Shield
        quantity = 5 if item.name == "Hazard Shield" else 1
        return resource_database.get_item(base_resource), quantity

    return PickupEntry(
        name=item.name,
        progression=tuple(
            _create_resources(progression)
            for progression in item.progression
        ),
        extra_resources=tuple(extra_resources),
        model=PickupModel(
            game=resource_database.game_enum,
            name=item.model_name,
        ),
        item_category=item.item_category,
        broad_category=item.broad_category,
        probability_offset=item.probability_offset,
        probability_multiplier=item.probability_multiplier,
        unlocks_resource=item.unlocks_ammo,
        respects_lock=ammo_requires_major_item,
        resource_lock=ammo.create_resource_lock(resource_database) if ammo is not None else None,
    )


def create_ammo_expansion(ammo: Ammo,
                          ammo_count: Tuple[int, ...],
                          requires_major_item: bool,
                          resource_database: ResourceDatabase,
                          ) -> PickupEntry:
    """
    Creates a Pickup for an expansion of the given ammo.
    :param ammo:
    :param ammo_count:
    :param requires_major_item:
    :param resource_database:
    :return:
    """
    resources = [(resource_database.get_item(item), count)
                 for item, count in zip(ammo.items, ammo_count)]

    if resource_database.item_percentage is not None:
        resources.append((resource_database.item_percentage, 1))

    return PickupEntry(
        name=ammo.name,
        progression=(),
        extra_resources=tuple(resources),
        model=PickupModel(
            game=resource_database.game_enum,
            name=ammo.model_name,
        ),
        item_category=ammo.item_category,
        broad_category=ammo.broad_category,
        respects_lock=requires_major_item,
        resource_lock=ammo.create_resource_lock(resource_database),
    )


def create_dark_temple_key(key_number: int,
                           temple_index: int,
                           resource_database: ResourceDatabase,
                           ) -> PickupEntry:
    """
    Creates a Dark Temple Key
    :param key_number:
    :param temple_index: The index of the temple: Dark Agon, Dark Torvus, Hive Temple
    :param resource_database:
    :return:
    """
    TEMPLE_KEY_CATEGORY = ItemCategory(
        name="temple_key",
        long_name="",
        hint_details=("a ", "red Temple Key"),
        is_major=False,
        is_key=True
    )

    return PickupEntry(
        name=echoes_items.DARK_TEMPLE_KEY_NAMES[temple_index].format(key_number + 1),
        progression=((resource_database.get_item(echoes_items.DARK_TEMPLE_KEY_ITEMS[temple_index][key_number]), 1),),
        model=PickupModel(
            game=resource_database.game_enum,
            name=echoes_items.DARK_TEMPLE_KEY_MODEL,
        ),
        item_category=TEMPLE_KEY_CATEGORY,
        broad_category=GENERIC_KEY_CATEGORY,
        probability_offset=3,
    )


def create_sky_temple_key(key_number: int,
                          resource_database: ResourceDatabase,
                          ) -> PickupEntry:
    """

    :param key_number:
    :param resource_database:
    :return:
    """
    SKY_TEMPLE_KEY_CATEGORY = ItemCategory(
        name="sky_temple_key",
        long_name="",
        hint_details=("a ", "Sky Temple Key"),
        is_major=False,
        is_key=True
    )

    return PickupEntry(
        name="Sky Temple Key {}".format(key_number + 1),
        progression=((resource_database.get_item(echoes_items.SKY_TEMPLE_KEY_ITEMS[key_number]), 1),),
        model=PickupModel(
            game=resource_database.game_enum,
            name=echoes_items.SKY_TEMPLE_KEY_MODEL,
        ),
        item_category=SKY_TEMPLE_KEY_CATEGORY,
        broad_category=GENERIC_KEY_CATEGORY,
        probability_offset=3,
    )


def create_energy_cell(cell_index: int,
                       resource_database: ResourceDatabase,
                       ) -> PickupEntry:
    ENERGY_CELL_CATEGORY = ItemCategory(
        name="energy_cell",
        long_name="",
        hint_details=("an ", "energy cell"),
        is_major=False,
        is_key=True
    )

    return PickupEntry(
        name=f"Energy Cell {cell_index + 1}",
        progression=(
            (resource_database.get_item(corruption_items.ENERGY_CELL_ITEMS[cell_index]), 1),
        ),
        extra_resources=(
            (resource_database.get_item(corruption_items.ENERGY_CELL_TOTAL_ITEM), 1),
            (resource_database.item_percentage, 1),
        ),
        model=PickupModel(
            game=resource_database.game_enum,
            name=corruption_items.ENERGY_CELL_MODEL,
        ),
        item_category=ENERGY_CELL_CATEGORY,
        broad_category=GENERIC_KEY_CATEGORY,
        probability_offset=0.25,
    )


def create_artifact(artifact_index: int,
                    minimum_progression: int,
                    resource_database: ResourceDatabase,
                    ) -> PickupEntry:
    ARTIFACT_CATEGORY = ItemCategory(
        name="artifact",
        long_name="",
        hint_details=("an ", "artifact"),
        is_major=False,
        is_key=True
    )

    return PickupEntry(
        name=prime_items.ARTIFACT_NAMES[artifact_index],
        progression=(
            (resource_database.get_item(prime_items.ARTIFACT_ITEMS[artifact_index]), 1),
        ),
        model=PickupModel(
            game=resource_database.game_enum,
            name=prime_items.ARTIFACT_MODEL[artifact_index],
        ),
        item_category=ARTIFACT_CATEGORY,
        broad_category=GENERIC_KEY_CATEGORY,
        probability_offset=0.25,
        required_progression=minimum_progression,
    )


def create_echoes_useless_pickup(resource_database: ResourceDatabase) -> PickupEntry:
    """
    Creates an Energy Transfer Module pickup.
    :param resource_database:
    :return:
    """
    return PickupEntry(
        name="Energy Transfer Module",
        progression=(
            (resource_database.get_item(echoes_items.USELESS_PICKUP_ITEM), 1),
        ),
        model=PickupModel(
            game=resource_database.game_enum,
            name=echoes_items.USELESS_PICKUP_MODEL,
        ),
        item_category=USELESS_ITEM_CATEGORY,
        broad_category=USELESS_ITEM_CATEGORY,
    )


def create_nothing_pickup(resource_database: ResourceDatabase) -> PickupEntry:
    """
    Creates a Nothing pickup.
    :param resource_database:
    :return:
    """
    return PickupEntry(
        name="Nothing",
        progression=(
            (resource_database.get_item_by_name("Nothing"), 1),
        ),
        model=PickupModel(
            game=resource_database.game_enum,
            name="Nothing",
        ),
        item_category=USELESS_ITEM_CATEGORY,
        broad_category=USELESS_ITEM_CATEGORY,
    )


def create_visual_etm() -> PickupEntry:
    """
    Creates an ETM that should only be used as a visual pickup.
    :return:
    """
    return PickupEntry(
        name="Unknown item",
        progression=tuple(),
        model=PickupModel(
            game=RandovaniaGame.METROID_PRIME_ECHOES,
            name=echoes_items.USELESS_PICKUP_MODEL,
        ),
        item_category=USELESS_ITEM_CATEGORY,
        broad_category=USELESS_ITEM_CATEGORY,
    )



from dataclasses import dataclass
from typing import Iterator, Tuple, FrozenSet, List, Callable, TypeVar, Type

from randovania.bitpacking import bitpacking
from randovania.bitpacking.bitpacking import BitPackValue, BitPackDecoder
from randovania.game_description import default_database
from randovania.game_description.world.area import Area
from randovania.game_description.world.area_location import AreaLocation
from randovania.games.game import RandovaniaGame


def area_locations_with_filter(game: RandovaniaGame, condition: Callable[[Area], bool]) -> List[AreaLocation]:
    world_list = default_database.game_description_for(game).world_list
    areas = [
        AreaLocation(world.world_asset_id, area.area_asset_id)
        for world in world_list.worlds
        for area in world.areas
        if condition(area)
    ]
    return list(sorted(areas))


T = TypeVar("T")
SelfType = TypeVar("SelfType")


@dataclass(frozen=True)
class LocationList(BitPackValue):
    locations: FrozenSet[AreaLocation]
    game: RandovaniaGame

    @classmethod
    def areas_list(cls, game: RandovaniaGame) -> List[AreaLocation]:
        return area_locations_with_filter(game, lambda area: True)

    @classmethod
    def element_type(cls) -> Type[AreaLocation]:
        return AreaLocation

    @classmethod
    def with_elements(cls: Type[SelfType], elements: Iterator[AreaLocation], game: RandovaniaGame) -> SelfType:
        elements_set = frozenset(sorted(elements))
        all_locations = frozenset(cls.areas_list(game))
        return cls(elements_set & all_locations, game)

    def bit_pack_encode(self, metadata) -> Iterator[Tuple[int, int]]:
        areas = self.areas_list(self.game)
        locations = list(sorted(self.locations))
        yield from bitpacking.pack_sorted_array_elements(locations, areas)

    @classmethod
    def bit_pack_unpack(cls, decoder: BitPackDecoder, metadata) -> "LocationList":
        game = metadata["reference"].game
        return cls.with_elements(bitpacking.decode_sorted_array_elements(decoder, cls.areas_list(game)), game)

    @property
    def as_json(self) -> List[dict]:
        return [location.as_json for location in self.locations]

    @classmethod
    def from_json(cls, value: List[dict], game: RandovaniaGame) -> "LocationList":
        if not isinstance(value, list):
            raise ValueError("StartingLocation from_json must receive a list, got {}".format(type(value)))

        elements = [cls.element_type().from_json(location) for location in value]
        return cls.with_elements(elements, game)

    def ensure_has_location(self: SelfType, area_location: AreaLocation, enabled: bool) -> SelfType:
        new_locations = set(self.locations)

        if enabled:
            new_locations.add(area_location)
        elif area_location in new_locations:
            new_locations.remove(area_location)

        return self.with_elements(iter(new_locations), self.game)

    def ensure_has_locations(self: SelfType, area_locations: List[AreaLocation], enabled: bool) -> SelfType:
        new_locations = set(self.locations)

        for area_location in area_locations:
            if enabled:
                new_locations.add(area_location)
            elif area_location in new_locations:
                new_locations.remove(area_location)

        return self.with_elements(iter(new_locations), self.game)
